"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT firstED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TypedDict


class DataRange(TypedDict):
    """
    Represents a time range with start and end dates.

    Attributes
    ----------
    started_at: str
        The start date of the range in ISO 8601 format.
    ended_at: str
        The end date of the range in ISO 8601 format.
    """
    started_at: str
    ended_at: str


class Extension(TypedDict):
    """
    Represents an extension with its details.

    Attributes
    ----------
    extension_id: str
        The unique identifier for the extension.
    URL: str
        The URL where the extension is hosted.
    type: str
        The type of the extension.
    date_range: DataRange
        The date range during which the extension data is relevant.
    """
    extension_id: str
    URL: str
    type: str
    date_range: DataRange


class Game(TypedDict):
    """
    Represents a game with its details.

    Attributes
    ----------
    game_id: str
        The unique identifier for the game.
    URL: str
        The URL where the game is hosted or related to.
    type: str
        The type of the game.
    date_range: DataRange
        The date range during which the game data is relevant.
    """
    game_id: str
    URL: str
    type: str
    date_range: DataRange



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING
import datetime
import logging
import json
import time

if TYPE_CHECKING:
    from typing import Any, Union, Dict, Optional
    import aiohttp

__all__ = ('setup_logging', 'json_or_text', 'convert_rfc3339', 'datetime_to_str', 'ExponentialBackoff')


def setup_logging(*,
                  handler: Optional[logging.Handler] = None,
                  level: Optional[int] = None,
                  root: bool = True) -> None:
    """
    Setup logging configuration.
    """
    if level is None:
        level = logging.INFO

    if handler is None:
        handler = logging.StreamHandler()

    dt_fmt = '%Y-%m-%d %H:%M:%S'
    formatter = logging.Formatter('[{asctime}] [{levelname}] {name}: {message}', dt_fmt, style='{')

    if root:
        logger = logging.getLogger()
    else:
        library, _, _ = __name__.partition('.')
        logger = logging.getLogger(library)

    handler.setFormatter(formatter)
    logger.setLevel(level)
    logger.addHandler(handler)


async def json_or_text(response: aiohttp.ClientResponse) -> Union[Dict[str, Any], str]:
    """Read response from aiohttp.ClientResponse, parse as JSON if content-type is 'application/json',
    otherwise return response text."""
    text = await response.text(encoding='utf-8')
    try:
        if 'application/json' in response.headers['content-type']:
            return json.loads(text)
    except KeyError:
        pass
    return text


def convert_rfc3339(timestamp: Optional[str]) -> Optional[datetime]:
    """
    Convert RFC3339 timestamp string to a datetime object (UTC +0).
    """
    if timestamp.endswith('Z'):
        timestamp = timestamp[:-1] + '+00:00'
    return None if (not timestamp) else datetime.datetime.fromisoformat(timestamp)


def datetime_to_str(__time: Optional[datetime], /) -> Optional[str]:
    """
    Convert local datetime object to UTC formatted RFC3339 timestamp string.
    """
    return None if time is None else __time.astimezone(datetime.timezone.utc).isoformat()


class ExponentialBackoff:
    """
    Handles retry intervals with exponential backoff.

    Parameters
    ----------
    base_delay: int
        The initial delay in seconds. The delay starts at this value and increases
        exponentially with each retry.
    max_delay: int
        The maximum delay between retries. The exponential increase is capped
        at this value.
    reset_interval: int
        The period in seconds after which the retry count is reset if no errors occur.
    """

    __slots__ = ('base_delay', 'max_delay', 'reset_interval', 'retry_count', 'last_failure_time')

    def __init__(self, base_delay: int = 1, max_delay: int = 180, reset_interval: int = 300) -> None:
        self.base_delay: int = base_delay
        self.max_delay: int = max_delay
        self.reset_interval: int = reset_interval
        self.retry_count: int = 0
        self.last_failure_time: float = time.monotonic()

    def get_delay(self) -> int:
        """

        Determine the delay before the next retry attempt.

        Returns
        -------
        int
            The delay in seconds before the next retry attempt.
        """
        current_time = time.monotonic()
        elapsed_time = current_time - self.last_failure_time

        if elapsed_time > self.reset_interval:
            self.retry_count = 0

        delay = min(self.base_delay * 2 ** self.retry_count, self.max_delay)
        self.retry_count += 1
        self.last_failure_time = current_time
        return delay



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT firstED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from .users import SpecificUser, Broadcaster
from typing import TypedDict, TYPE_CHECKING

if TYPE_CHECKING:
    from typing import List, Literal, NotRequired, Optional


# HypeTrain
class HypeTrainContributor(TypedDict):
    """
    Represents a contributor to a Hype Train event.

    Attributes
    ----------
    type: Literal['bits', 'subscription', 'other']
        The type of contribution (e.g., bits, subscription).
    user: str
        The user who made the contribution.
    total: int
        The total amount contributed by the user.
    """
    type: Literal['bits', 'subscription', 'other']
    user: str
    total: int


class HypeTrain(TypedDict):
    """
    Represents a Hype Train event.

    Attributes
    ----------
    id: str
        The unique identifier for the Hype Train.
    total: int
        The total amount of contributions.
    level: int
        The level achieved in the Hype Train.
    started_at: str
        The timestamp when the Hype Train started.
    goal: str
        The goal required to reach the next level.
    expires_at: str
        The timestamp when the Hype Train expires.
    broadcaster_id: str
        The ID of the broadcaster during the Hype Train.
    cooldown_end_time: str
        The timestamp when the cooldown period ends.
    last_contribution: HypeTrainContributor
        The last contribution made during the Hype Train.
    top_contributions: List[HypeTrainContributor]
        A list of the top contributions made during the Hype Train.
    """
    id: str
    total: int
    level: int
    started_at: str
    goal: str
    expires_at: str
    broadcaster_id: str
    cooldown_end_time: str
    last_contribution: HypeTrainContributor
    top_contributions: List[HypeTrainContributor]


# Polls
PollStatus = Literal['ACTIVE', 'COMPLETED', 'TERMINATED', 'ARCHIVED', 'MODERATED', 'INVALID']


class PollChoice(TypedDict, total=False):
    """
    Represents a choice in a poll.

    Attributes
    ----------
    id: str
        The unique identifier for the poll choice.
    title: str
        The title of the poll choice.
    votes: int
        The number of votes this choice received.
    bits_votes: int
        The number of bits votes this choice received.
    channel_points_votes: int
        The number of channel points votes this choice received.
    """
    id: str
    title: str
    votes: NotRequired[int]
    bits_votes: NotRequired[int]
    channel_points_votes: NotRequired[int]
    

class Poll(Broadcaster):
    """
    Represents a poll.

    Attributes
    ----------
    id: str
        The unique identifier for the poll.
    title: str
        The title of the poll.
    choices: List[PollChoice]
        A list of choices available in the poll.
    started_at: str
        The timestamp when the poll started.
    status: PollStatus
        The current status of the poll.
    duration: int
        The duration of the poll in seconds.
    ended_at: Optional[str]
        The timestamp when the poll ended, if applicable.
    bits_per_vote: int
        The cost of a vote in bits, if bits voting is enabled.
    bits_voting_enabled: bool
        Indicates if bits voting is enabled.
    channel_points_per_vote: int
        The cost of a vote in channel points, if channel points voting is enabled.
    channel_points_voting_enabled: bool
        Indicates if channel points voting is enabled.
    """
    id: str
    title: str
    choices: List[PollChoice]
    started_at: str
    status: PollStatus
    duration: int
    ended_at: Optional[str]
    bits_per_vote: int
    bits_voting_enabled: bool
    channel_points_per_vote: int
    channel_points_voting_enabled: bool


# Prediction
class Predictor(SpecificUser):
    """
    Represents a predictor in a channel prediction.

    Attributes
    ----------
    channel_points_won: Optional[int]
        The number of channel points won by the predictor.
    channel_points_used: int
        The number of channel points used by the predictor.
    """
    channel_points_won: Optional[int]
    channel_points_used: int


class Outcome(TypedDict, total=False):
    """
    Represents an outcome in a channel prediction.

    Attributes
    ----------
    id: str
        The unique identifier for the outcome.
    color: str
        The color associated with the outcome.
    title: str
        The title of the outcome.
    users: int
        The number of users who predicted this outcome.
    channel_points: int
        The total number of channel points used for this outcome.
    top_predictors: List[Predictor]
        A list of the top predictors for this outcome.
    """
    id: str
    color: str
    title: str
    users: NotRequired[int]
    channel_points: NotRequired[int]
    top_predictors: NotRequired[List[Predictor]]


class Prediction(TypedDict):
    """
    Represents a channel prediction.

    Attributes
    ----------
    id: str
        The unique identifier for the prediction.
    title: str
        The title of the prediction.
    outcomes: List[Outcome]
        A list of possible outcomes for the prediction.
    status: Literal['ACTIVE', 'CANCELED', 'LOCKED', 'RESOLVED']
        The current status of the prediction.
    ended_at: Optional[str]
        The timestamp when the prediction ended, if applicable.
    locked_at: Optional[str]
        The timestamp when the prediction was locked, if applicable.
    created_at: str
        The timestamp when the prediction was created.
    prediction_window: int
        The time window in seconds during which users can participate in the prediction.
    winning_outcome_id: Optional[str]
        The ID of the winning outcome, if the prediction has been resolved.
    """
    id: str
    title: str
    outcomes: List[Outcome]
    status: Literal['ACTIVE', 'CANCELED', 'LOCKED', 'RESOLVED']
    ended_at: Optional[str]
    locked_at: Optional[str]
    created_at: str
    prediction_window: int
    winning_outcome_id: Optional[str]


# Reward
class RewardImages(TypedDict):
    """
    Represents images for a channel reward.

    Attributes
    ----------
    url_1x: str
        The URL for the 1x resolution image.
    url_2x: str
        The URL for the 2x resolution image.
    url_4x: str
        The URL for the 4x resolution image.
    """
    url_1x: str
    url_2x: str
    url_4x: str


class BaseReward(TypedDict):
    """
    A base class used to Represents common attributes of a channel reward.

    Attributes
    ----------
    id: str
        The unique identifier for the reward.
    cost: int
        The cost of the reward in channel points.
    title: str
        The title of the reward.
    prompt: str
        The prompt or description of the reward.
    """
    id: str
    cost: int
    title: str
    prompt: str


class MaxPerStreamSetting(TypedDict):
    """
    Represents settings for the maximum number of times a reward can be redeemed per stream.

    Attributes
    ----------
    is_enabled: bool
        Indicates if the setting is enabled.
    max_per_stream: int
        The maximum number of times the reward can be redeemed per stream.
    """
    is_enabled: bool
    max_per_stream: int


class MaxPerUserPerStreamSetting(TypedDict):
    """
    Represents settings for the maximum number of times a reward can be redeemed per user per stream.

    Attributes
    ----------
    is_enabled: bool
        Indicates if the setting is enabled.
    max_per_user_per_stream: int
        The maximum number of times the reward can be redeemed per user per stream.
    """
    is_enabled: bool
    max_per_user_per_stream: int


class GlobalCooldownSetting(TypedDict):
    """
    Represents settings for the global cooldown of a reward.

    Attributes
    ----------
    is_enabled: bool
        Indicates if the global cooldown is enabled.
    global_cooldown_seconds: int
        The number of seconds for the global cooldown.
    """
    is_enabled: bool
    global_cooldown_seconds: int


class Reward(Broadcaster, BaseReward):
    """
    Represents a channel reward.

    Attributes
    ----------
    image: Optional[RewardImages]
        The custom image for the reward, if applicable.
    is_paused: bool
        Indicates if the reward is paused.
    is_enabled: bool
        Indicates if the reward is enabled.
    is_in_stock: bool
        Indicates if the reward is in stock.
    default_image: RewardImages
        The default image for the reward.
    background_color: str
        The background color for the reward.
    cooldown_expires_at: Optional[str]
        The timestamp when the cooldown for the reward expires, if applicable.
    is_user_input_required: bool
        Indicates if user input is required to redeem the reward.
    max_per_stream_setting: MaxPerStreamSetting
        The settings for the maximum number of redemptions per stream.
    global_cooldown_setting: GlobalCooldownSetting
        The settings for the global cooldown of the reward.
    max_per_user_per_stream_setting: MaxPerUserPerStreamSetting
        The settings for the maximum number of redemptions per user per stream.
    redemptions_redeemed_current_stream: Optional[int]
        The number of redemptions of the reward during the current stream, if applicable.
    should_redemptions_skip_request_queue: bool
        Indicates if redemptions should skip the request queue.
    """
    image: Optional[RewardImages]
    is_paused: bool
    is_enabled: bool
    is_in_stock: bool
    default_image: RewardImages
    background_color: str
    cooldown_expires_at: Optional[str]
    is_user_input_required: bool
    max_per_stream_setting: MaxPerStreamSetting
    global_cooldown_setting: GlobalCooldownSetting
    max_per_user_per_stream_setting: MaxPerUserPerStreamSetting
    redemptions_redeemed_current_stream: Optional[int]
    should_redemptions_skip_request_queue: bool


class RewardRedemption(Broadcaster, SpecificUser):
    """
    Represents a reward redemption.

    Attributes
    ----------
    id: str
        The unique identifier for the redemption.
    status: Literal['CANCELED', 'FULFILLED', 'UNFULFILLED']
        The current status of the redemption.
    reward: BaseReward
        The reward associated with this redemption.
    user_input: str
        The input provided by the user when redeeming the reward.
    redeemed_at: str
        The timestamp when the reward was redeemed.
    """
    id: str
    status: Literal['CANCELED', 'FULFILLED', 'UNFULFILLED']
    reward: BaseReward
    user_input: str
    redeemed_at: str




"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT firstED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import textwrap


class Template:
    """
    Represents a template for generating overlay HTML with dynamic content.
    """
    __slots__ = ()

    def __init__(self):
        pass

    def _index(self) -> str:
        """Generate the HTML template for the overlay."""
        index = '''
        <!DOCTYPE html>
        <html lang="en">
        <head>
            <meta charset="UTF-8">
            <meta name="viewport" content="width=device-width, initial-scale=1.0">
            <title>Overlay</title>
            <meta name="copyright" content="MIT License (c) 2024 by Snifo">
            <meta name="description" content="Overlay extension for twitch.py.">
            <meta name="author" content="Snifo">
            <meta name="generator" content="twitch.py">
            <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/animate.css/4.1.1/animate.min.css">
            <style> {styles} </style>
        </head>
        <body>
            <div id="alert-container" class="alert-container">
                <div id="alert-image-wrapper" class="alert-image-wrapper hide">
                    <img src id="alert-image" class="alert-image"> 
                </div>
                <div id="alert-text-wrapper" class="alert-text-wrapper">
                    <div id="alert-message" class="alert-message">
                    </div>
                </div>
            </div>
            <script>
            {filter_script} 
             </script>
            <script src="https://cdnjs.cloudflare.com/ajax/libs/howler/2.2.3/howler.min.js"></script>
            <script>
            {alert_script}
            </script>
        </body>
        </html>
        '''
        html = index.format(styles=self._styles,
                            filter_script=self._js_filter,
                            alert_script=self._js_alert)
        return html.replace('        ', '')

    @property
    def _styles(self) -> str:
        """Define the CSS styles for the overlay."""
        return '''
        :root {
            --animate-duration: 1.8s;
        }
        
        html, body {
            margin: 0;
            padding: 0;
            width: 100%;
            height: 100%;
            background: transparent;
            overflow: hidden;
        }
        
        .alert-container, .alert-warp {
            display: flex;
            flex-direction: column;
            width: 100%;
            height: 100%;
        }
        
        .alert-image-wrapper.hide {
            display: none;
        }
        
        /* Image */
        .alert-image-wrapper {
            display: flex;
            justify-content: center;
            align-items: center;
            flex: 7.8;
            overflow: hidden;
        }
        
        .alert-image {
            width: 100%;
            height: auto;
            max-width: 100%;
            max-height: 100%;
            object-fit: contain;
            display: block;
        }
        
        .alert-image.hidden {
            visibility: hidden;
        }
        
        /* Text */
        .alert-text-wrapper {
            flex: 2.2; /* 20% height */
            display: flex;
            justify-content: center;
            align-items: center;
            overflow: hidden;
            position: relative;
            width: 100%;
            padding: 0;
            box-sizing: border-box;
        }
        
        .alert-message {
            font-size: 64px;
            color: rgb(255, 255, 255); /* White text */
            font-family: sans-serif;
            font-weight: 400;
            text-align: center;
            line-height: 1.2;
            letter-spacing: 1px;
            text-shadow: 2px 2px 2px rgba(0, 0, 0, 1); /* Dark, sharp shadow */
            padding: 0 20px;
            box-sizing: border-box;
            width: 100%;
            white-space: normal; /* Allow text to wrap */
            word-break: break-word; /* Prevent text overflow */
        }
        
        /* Highlighted text styling */
        .alert-message .highlight {
            color: cyan;
            font-weight: bold; /* Example highlight style */
            display: inline-block; /* Ensure it's an inline element */
        }
        
        /* Small text styling */
        .alert-message .small-text {
            font-size: 0.6em; /* Reduce font size to 60% */
            display: inline-block;
        }
        '''

    @property
    def _js_filter(self) -> str:
        """JavaScript to set the default filter for the overlay."""
        return '''
        // Set default filter if not present
        const urlParams = new URLSearchParams(window.location.search);
        const filterParam = urlParams.get('default') || 'all';

        if (!urlParams.has('default')) {
            urlParams.set('default', 'all');
            window.history.replaceState({}, '', `${window.location.pathname}?${urlParams}`);
        }
        '''

    @property
    def _js_alert(self) -> str:
        """JavaScript for handling alerts, including animations, fonts, and images."""
        js_template = '''
        let reconnectInterval = 5000; // Time before attempting to reconnect
        const maxReconnectInterval = 60000; // Maximum delay for reconnection
        let isAnimating = false;
        let Queue = [];

        {initialize_audio}

        {alert_animate}

        {load_google_font}

        {load_image}

        {alert_update}

        {process_queue}

        {connect_websocket}

        connectWebSocket();         
        '''

        js_script = js_template.format(
            initialize_audio=self._js_initialize_audio,
            alert_animate=self._js_alert_animate,
            load_google_font=self._js_load_google_font,
            load_image=self._js_load_image,
            alert_update=self._js_alert_update,
            process_queue=self._js_process_queue,
            connect_websocket=self._js_connect_websocket
        )
        return js_script

    @property
    def _js_initialize_audio(self) -> str:
        """JavaScript function to initialize audio with Howler.js."""
        return '''
        const Initializeaudio = (url) => {
            return new Promise((resolve, reject) => {
                const audio = new Howl({
                    src: [url],
                    autoplay: false,
                    volume: 1.0,
                    onload: () => resolve(audio), // Resolve when the audio is loaded
                    onloaderror: () => reject(new Error('Failed to load audio')) // Reject if audio loading fails
                });
            });
        };
        '''

    @property
    def _js_alert_animate(self) -> str:
        """JavaScript function to animate an element using Animate.Style."""
        return '''
        // Animate an element using Animate.css
        const AlertAnimation = (element, animation) => {
            return new Promise((resolve) => {
                element.classList.add('animate__animated', `animate__${animation}`);
        
                // Resolve when the animation ends
                function handleAnimationEnd(event) {
                    event.stopPropagation();
                    element.classList.remove('animate__animated', `animate__${animation}`);
                    resolve('Animation ended');
                }
                element.addEventListener('animationend', handleAnimationEnd, { once: true });
            });
        };      
        '''

    @property
    def _js_load_google_font(self) -> str:
        """JavaScript function to load Google Fonts dynamically."""
        return '''
        // Load Google Fonts dynamically
        const LoadGoogleFonts = (fontName) => {
            return new Promise((resolve, reject) => {
                const link = document.createElement('link');
                link.href = `https://fonts.googleapis.com/css2?family=${fontName.replace(/ /g, '+')}&display=swap`;
                link.rel = 'stylesheet';
                link.onload = resolve; // Resolve when the font is loaded
                link.onerror = () => reject(new Error('Failed to load font')); // Reject if font loading fails
                document.head.appendChild(link);
            });
        };
        '''

    @property
    def _js_load_image(self) -> str:
        """JavaScript function to load an image and set its source."""
        return '''
        const LoadImage = (element, url) => {
            return new Promise((resolve, reject) => {
                element.src = url;  // Set the image source URL
                element.onload = () => {
                    resolve();  // Resolve the promise when the image is successfully loaded
                };
                element.onerror = () => {
                    reject(new Error('Failed to load image')); 
                };
            });
};
        '''

    @property
    def _js_alert_update(self) -> str:
        """JavaScript function to update the alert with new content and animations."""
        return '''
        // Main function to update the alert with dynamic content
        function Update(data) {
            return new Promise(async (resolve) => {
                const imageWarpElement = document.getElementById('alert-image-wrapper');
                const imageElement = document.getElementById('alert-image');
                const textElement = document.getElementById('alert-message');
                const alertWarpElement = document.getElementById('alert-container');
        
                if (!textElement || !alertWarpElement) {
                    console.error('Required element(s) not found.');
                    return resolve(); // Resolve immediately if elements are not found
                }
        
                // Load and apply custom font
                try {
                    await LoadGoogleFonts(data.font_name);
                    textElement.style.fontFamily = `'${data.font_name}', sans-serif`;
                } catch (error) {
                    textElement.style.fontFamily = 'sans-serif'; // Fallback to default font
                    console.error(error.message);
                }
        
                
                textElement.style.fontSize = `${data.font_size}px`
                
                textElement.style.color = data.text_color
        
                // Load and play audio if available
                let audio = null;
                if (data.audio) {
                    try {
                        audio = await Initializeaudio(data.audio);
                    } catch (error) {
                        console.error('Error initializing audio:', error);
                    }
                }
        
                // Apply text formatting (highlight and small text)
                const fontSizeRegex = /%([^%]+)%/g;
                const highlightRegex = /<<([^<>]+)>>/g;
                let formattedText = data.text
                    .replace(fontSizeRegex, '<div class="small-text">$1</div>')
                    .replace(highlightRegex, `<div id="highlight" class="highlight">$1</div>`);
                
        
                // Play audio if loaded
                if (audio) {
                    audio.play();
                }
        
                // Load and display image if available
                imageWarpElement.classList.add('hide');
                
                if (data.image) {
                    try {
                        await LoadImage(imageElement, data.image);
                        imageWarpElement.classList.remove('hide');
                    } catch (error) {
                        console.error(error.message);
                    }
                }
                
                textElement.innerHTML = formattedText;
        
                // Animate highlighted text if present
                const highlightElements = document.querySelectorAll('#highlight');
        
                if (highlightElements.length > 0) {
                    highlightElements.forEach((element) => {
                        // Add the animation class to each element
                        element.style.color = data.text_highlight_color
                        element.classList.add('animate__animated', `animate__${data.text_animation}`,
                        'animate__infinite');
                    });
                }
        
                // Animate the alert and manage its lifecycle
                AlertAnimation(alertWarpElement, data.start_animation).then(() => {
                    setTimeout(() => {
                        AlertAnimation(alertWarpElement, data.end_animation).then(() => {
                            // Clean up: reset elements after the animation ends
                            imageElement.src = '';
                            textElement.innerHTML = '';
                            textElement.style = '';
        
                            // Stop audio if playing
                            if (audio && audio.playing()) {
                                audio.stop();
                            }
                            resolve(); // Resolve the promise when everything is done
                        });
                    }, data.alert_duration * 1000); // Wait for alert duration before ending
                });
            });
        }
        '''

    @property
    def _js_process_queue(self) -> str:
        """JavaScript function to process a queue of alerts with a delay."""
        return '''
        const processQueue = () => {
            if (Queue.length > 0 && !isAnimating) {
                const data = Queue.shift();

                // Check if data is a non-empty object
                if (data && Object.keys(data).length > 0) {
                    isAnimating = true;
                    Update(data)
                        .then(() => {
                            isAnimating = false;
                            setTimeout(processQueue, 300);
                        })
                        .catch(error => {
                            // Log and handle errors, reset isAnimating and continue processing
                            console.error('Error in Update:', error);
                            isAnimating = false;
                            setTimeout(processQueue, 300);
                        });
                } else {
                    // Data is empty or invalid, just reset isAnimating and continue processing
                    isAnimating = false;
                    setTimeout(processQueue, 300);
                }
            }
        };
        '''

    @property
    def _js_connect_websocket(self) -> str:
        """JavaScript function to connect to a WebSocket server and handle messages."""
        return '''
        function connectWebSocket() {
            try {
                const urlParams = new URLSearchParams(window.location.search);
                const filterParam = urlParams.get('default') || 'all';
                
                const host = window.location.hostname;
                const port = window.location.port;
                
                ws = new WebSocket(`ws://${host}:${port}/ws?default=${encodeURIComponent(filterParam)}`);
        
                ws.onopen = () => {
                    console.log('WebSocket connection established');
                    reconnectInterval = 5000; // Reset reconnection interval
                };
        
                ws.onmessage = (event) => {
                    try {
                        const data = JSON.parse(event.data);
                        console.log('WebSocket message received:', data);
        
                        Queue.push(data);
                        processQueue();
                    } catch (err) {
                        console.error('Error processing WebSocket message:', err);
                    }
                };
        
                ws.onclose = (event) => {
                    console.warn(`WebSocket closed: ${event.reason || 'No reason provided'}`);
                    scheduleReconnect();
                };
        
                ws.onerror = (error) => {
                    console.error('WebSocket error:', error);
                    ws.close(); // Close and trigger onclose to handle reconnection
                };
        
            } catch (err) {
                console.error('WebSocket initialization failed:', err);
                scheduleReconnect();
            }
        }
        
        
        function scheduleReconnect() {
            console.log(`Reconnecting in ${reconnectInterval / 1000} seconds...`);
            setTimeout(() => {
                reconnectInterval = Math.min(reconnectInterval * 2, maxReconnectInterval); // Exponential backoff
                connectWebSocket();
            }, reconnectInterval);
        }
        '''

    def __str__(self) -> str:
        return textwrap.dedent(self._index()).strip()



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT firstED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING
from .template import Template
from aiohttp import web

if TYPE_CHECKING:
    from .attachment import Attachment
    from .geteway import WebSocket
    from typing import Optional

import logging
_logger = logging.getLogger(__name__)


class Route:
    """
    Handles HTTP and WebSocket routes for the application.
    """
    __slots__ = ('gateway', 'attachment', 'default_template')

    def __init__(self, router: web.Application.router, gateway: WebSocket, attachment: Attachment):
        self.gateway: WebSocket = gateway
        self.attachment: Attachment = attachment
        router.add_get('/', self.handle_request)
        router.add_get('/ws', self.websocket_handler)
        router.add_get('/upload/{filename}', self.serve_file)

        # Loading template.
        self.default_template: str = str(Template())

    async def handle_request(self, request: web.Request) -> web.Response:
        """Handle HTTP requests to the main page."""
        response = web.Response(text=self.default_template, content_type='text/html')
        return response

    async def serve_file(self, request: web.Request) -> web.Response:
        """Serve uploaded files from the /upload endpoint."""
        filename: str = request.match_info['filename']
        file_content: Optional[bytes] = self.attachment.get_attachment(filename)
        if not file_content:
            _logger.warning('File not found: %s', filename)
            return web.Response(text='File not found', status=404)
        return web.Response(body=file_content, content_type='application/octet-stream')

    async def websocket_handler(self, request: web.Request) -> web.WebSocketResponse:
        """Handle WebSocket connections."""
        ws = await self.gateway.handle_connection(request)
        return ws



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT firstED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict
from .users import SpecificUser

if TYPE_CHECKING:
    from typing import List, Optional, Literal, NotRequired


class EmoteImages(TypedDict):
    """
    Represents the URLs for different sizes of emote images.

    Attributes
    ----------
    url_1x: str
        The URL for the emote image at 1x resolution.
    url_2x: str
        The URL for the emote image at 2x resolution.
    url_4x: str
        The URL for the emote image at 4x resolution.
    """
    url_1x: str
    url_2x: str
    url_4x: str


class Emote(TypedDict):
    """
    Represents an emote.

    Attributes
    ----------
    id: str
        The unique identifier for the emote.
    name: str
        The name of the emote.
    images: EmoteImages
        The URLs for different sizes of the emote images.
    template_url: str
        The URL template for the emote images.
    emote_type: str
        The type of the emote.
    emote_set_id: str
        The set ID to which the emote belongs.
    owner_id: str
        The ID of the emotes owner.
    format: Literal['animated', 'static']
        The format of the emote (animated or static).
    scale: Literal['1.0', '2.0', '3.0']
        The scale of the emote.
    theme_mode: Literal['dark', 'light']
        The theme mode for which the emote is designed (dark or light).
    """
    id: str
    name: str
    images: EmoteImages
    template_url: str
    emote_type: str
    emote_set_id: str
    owner_id: str
    format: Literal['animated', 'static']
    scale: Literal['1.0', '2.0', '3.0']
    theme_mode: Literal['dark', 'light']


class BadgeVersion(TypedDict):
    """
    Represents a specific version of a badge.

    Attributes
    ----------
    id: str
        The unique identifier for the badge version.
    image_url_1x: str
        The URL of the badge image at 1x resolution.
    image_url_2x: str
        The URL of the badge image at 2x resolution.
    image_url_4x: str
        The URL of the badge image at 4x resolution.
    title: str
        The title of the badge version.
    description: str
        A brief description of the badge version.
    click_action: Optional[str]
        The action that occurs when the badge is clicked.
    click_url: Optional[str]
        The URL to which the user is directed when the badge is clicked.
    """
    id: str
    image_url_1x: str
    image_url_2x: str
    image_url_4x: str
    title: str
    description: str
    click_action: Optional[str]
    click_url: Optional[str]


class Badge(TypedDict):
    """
    Represents a badge with multiple versions.

    Attributes
    ----------
    set_id: str
        The unique identifier for the badge set.
    versions: List[BadgeVersion]
        A list of the badge's versions.
    """
    set_id: str
    versions: List[BadgeVersion]


class Settings(TypedDict, total=False):
    """
    Represents various chat settings.

    Attributes
    ----------
    broadcaster_id: str
        The unique identifier for the broadcaster.
    emote_mode: bool
        Whether the emote-only mode is enabled.
    follower_mode: bool
        Whether the follower-only mode is enabled.
    follower_mode_duration: Optional[int]
        The duration in minutes for which a user must follow to chat.
    moderator_id: str
        The unique identifier for the moderator.
    non_moderator_chat_delay: NotRequired[bool]
        Whether there is a chat delay for non-moderators.
    non_moderator_chat_delay_duration: NotRequired[int]
        The duration of the chat delay for non-moderators.
    slow_mode: bool
        Whether the slow mode is enabled.
    slow_mode_wait_time: Optional[bool]
        The wait time in seconds for the slow mode.
    subscriber_mode: bool
        Whether the subscriber-only mode is enabled.
    unique_chat_mode: bool
        Whether the unique chat mode is enabled (prevents repeated messages).
    """
    broadcaster_id: str
    emote_mode: bool
    follower_mode: bool
    follower_mode_duration: Optional[int]
    moderator_id: str
    non_moderator_chat_delay: NotRequired[bool]
    non_moderator_chat_delay_duration: NotRequired[int]
    slow_mode: bool
    slow_mode_wait_time: Optional[bool]
    subscriber_mode: bool
    unique_chat_mode: bool


class MessageDropReason(TypedDict):
    """
    Represents the reason for a message drop.

    Attributes
    ----------
    code: str
        The error code associated with the message drop.
    message: str
        A description of the reason for the message drop.
    """
    code: str
    message: str


class SendMessageStatus(TypedDict, total=False):
    """
    Represents the status of a user sent message.

    Attributes
    ----------
    message_id: str
        The unique identifier for the message.
    is_sent: bool
        Whether the message was successfully sent.
    drop_reason: NotRequired[MessageDropReason]
        The reason for a message drop, if applicable.
    """
    message_id: str
    is_sent: bool
    drop_reason: NotRequired[MessageDropReason]


UserChatColors = Literal['blue', 'blue_violet', 'cadet_blue', 'chocolate', 'coral', 'dodger_blue', 'firebrick',
                         'golden_rod', 'green', 'hot_pink', 'orange_red', 'red', 'sea_green', 'spring_green',
                         'yellow_green']


class UserChatColor(SpecificUser):
    """
    Represents a user's chat color.

    Attributes
    ----------
    color: str
        The hexadecimal color code chosen by the user for their chat messages.
    """
    color: str



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT firstED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from .attachment import Attachment
from typing import TYPE_CHECKING
from .geteway import WebSocket
from .route import Route
from aiohttp import web

if TYPE_CHECKING:
    from typing import Dict, Any, Optional, Tuple
    from twitch import Client

import logging
_logger = logging.getLogger(__name__)


class Overlay:
    """
    Represents the overlay server application.

    This class sets up and manages the overlay server, including routing,
    handling WebSocket connections, managing file attachments, and sending alerts.

    ???+ Notes
        - If sound is not working in the web browser, make sure to set the website
          permissions for sound to "Allow."

        - To change the default template used by the server, you can set
          `route.default_template = 'Your HTML'` on setup.

    Parameters
    ----------
    client: Optional[Client]
        An optional `Client` instance for event dispatching.
    host: str
        The host address for the server (default is 'localhost').
    port: int
        The port number for the server (default is 5050).
    **options: Dict[str, Any]
        Additional options to configure the `web.Application` instance.
    """

    __slots__ = ('app', 'host', 'port', 'gateway', '_attachment', 'route', '_client')

    def __init__(self, client: Optional[Client] = None, host: str = 'localhost', port: int = 5050, **options) -> None:
        self.app = web.Application(**options)
        self.host = host
        self.port = port
        self.gateway = WebSocket()
        self._attachment = Attachment()
        self.route = Route(self.app.router, self.gateway, self._attachment)
        self._client: Optional[Client] = client

    def url(self, default: str = 'all') -> str:
        """
        Get the server URI.

        Parameters
        ----------
        default: str
            The filter parameter for the alert.

        Returns
        -------
        str
            The full URL of the server with the specified filter parameter.
        """
        return f'http://{self.host}:{self.port}/?default={default}'

    async def start_app(self, handle_signals: bool = False, **options) -> Tuple[web.AppRunner, web.TCPSite]:
        """
        Start the overlay server.

        Sets up the application runner and site, and starts the server.

        ???+ Note
            Call this method within your `setup_hook()`.

        Parameters
        ----------
        handle_signals: bool
            Whether to handle signals for graceful shutdown (default is False).
        **options
            Additional options to pass to the `web.TCPSite` constructor.

        Returns
        -------
        Tuple[web.AppRunner, web.TCPSite]
            The application runner and TCP site objects.
        """
        runner = web.AppRunner(self.app, handle_signals=handle_signals)
        await runner.setup()
        site = web.TCPSite(runner, self.host, self.port, **options)
        _logger.debug('Starting overlay server on http://%s:%s', self.host, self.port)
        await site.start()
        _logger.debug('Overlay server started successfully.')
        if self._client is not None:
            self._client.dispatch('overlay_ready')
        return runner, site

    async def alert(
            self,
            text: str,
            font_name: str = 'Roboto',
            font_size: float = 64.0,
            text_color: str = '#FFFFFF',
            text_highlight_color: str = '#6441a5',
            text_animation: str = 'pulse',
            start_animation: str = 'fadeIn',
            end_animation: str = 'fadeOut',
            image: Optional[str] = None,
            audio: Optional[str] = None,
            alert_duration: float = 5.0,
            default: str = 'all'):
        """
        Send an alert with specified parameters.

        ???+ Info
             The overlay uses [Animate.css](https://animate.style/).

             Be careful with animation names as they are **case-sensitive**.
    
        Parameters
        ----------
        text: str
            The text message to display.
        font_name: str
            The name of the font to use.
        font_size: float
            The size of the text font.
        text_color: str
            The color of the text.
        text_highlight_color: str
            The highlight color for text.
        text_animation: str
            The animation applied to the text.
            Refer to [Animate.css](https://animate.style/) for available text animations.
        start_animation: str
            The animation applied at the start of the alert.
            Refer to [Animate.css](https://animate.style/) for available start animations.
        end_animation: str
            The animation applied at the end of the alert.
            Refer to [Animate.css](https://animate.style/) for available end animations.
        image: Optional[str]
            The path to the image to display.
        audio: Optional[str]
            The path to the audio file to play.
        alert_duration: float
            The duration of the alert in seconds.
        default: str
            The `default` parameter specifies which filter key will trigger alerts.
            If set to a specific value (e.g., `'special'`), only alerts with that key will be processed.
            For example, if you set `default='special'`, alerts will only be triggered for the 'special' filter.
            To ensure the alert is processed, include this filter in the request URL.
            Example URL: `http://localhost:5050/?default=special`
        """
        data: Dict[str, Any] = {
            'text': text,
            'font_name': font_name,
            'font_size': font_size,
            'text_color': text_color,
            'text_highlight_color': text_highlight_color,
            'text_animation': text_animation,
            'start_animation': start_animation,
            'end_animation': end_animation,
            'image': self._attachment.get_path_key(image),
            'audio': self._attachment.get_path_key(audio),
            'alert_duration': alert_duration
        }
        await self.gateway.send_data(data, default)
        _logger.debug('Alert sent to default `%s` with data: %s', default, data)
        if self._client is not None:
            self._client.dispatch('overlay_alert', default, data)

    def add_attachment(self, name: str, path: str) -> None:
        """
        Add a new file attachment.

        Parameters
        ----------
        name: str
            The name of the attachment.
        path: str
            The file path of the attachment.
        """
        self._attachment.load_file_into_memory(name, path)

    def remove_attachment(self, name: str) -> None:
        """
        Remove an existing file attachment.

        Parameters
        ----------
        name: str
            The name of the attachment to remove.
        """
        self._attachment.remove_attachment(name)

    def clear_attachments(self) -> None:
        """
        Clear all attachments.
        """
        self._attachment.clear()

    def get_attachments(self) -> Dict[str, str]:
        """
        Retrieve all file attachments.

        Returns
        -------
        Dict[str, str]
            A dictionary where the keys represent attachment names and
            the values represent their corresponding upload paths.
        """
        return self._attachment.path_keys



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, overload
from .utils import datetime_to_str
import datetime

if TYPE_CHECKING:
    from typing import Optional, List, AsyncGenerator, Dict, Any, Literal
    from .types import Data, PData, channels, streams
    from .state import ConnectionState
    from .user import User

__all__ = ('Stream', 'BroadcasterStream')


class Stream:
    """
    Represents a channel stream.
    """

    __slots__ = ('_state', '_user_id', '_auth_user_id')

    def __init__(self, user_id: str, auth_user_id: str, *, state: ConnectionState) -> None:
        self._state: ConnectionState = state
        self._user_id: str = user_id
        self._auth_user_id: str = auth_user_id

    @overload
    async def get_live(self) -> streams.StreamInfo:
        ...

    @overload
    async def get_live(self) -> None:
        ...

    async def get_live(self) -> Optional[streams.StreamInfo]:
        """
        Retrieve the live stream information for the broadcaster.

        Returns
        -------
        Optional[streams.StreamInfo]
            A dictionary representing the live stream information if the broadcaster is live; otherwise, None.
        """
        data: PData[List[streams.StreamInfo]] = await self._state.http.get_streams(self._auth_user_id,
                                                                                   user_ids=[self._user_id])
        return data['data'][0] if len(data['data']) != 0 else None

    async def create_marker(self, description: Optional[str] = None) -> streams.StreamMarkerInfo:
        """
        Create a stream marker for the broadcaster.

        | Scopes                     | Description                                 |
        | -------------------------- | --------------------------------------------|
        | `channel:manage:broadcast` | Manage a channel’s broadcast configuration. |

        Parameters
        ----------
        description: Optional[str]
            A description for the marker.

        Returns
        -------
        streams.StreamMarkerInfo
            A dictionary representing the stream marker information.
        """
        data: Data[List[streams.StreamMarkerInfo]] = await self._state.http.create_stream_marker(self._auth_user_id,
                                                                                                 self._user_id,
                                                                                                 description)
        return data['data'][0] if len(data['data']) != 0 else None

    async def send_shoutout(self, user: User) -> None:
        """
        Send a shoutout to another broadcaster.

        | Scopes                       | Description                       |
        | ---------------------------- | ----------------------------------|
        | `moderator:manage:shoutouts` | Manage a broadcaster’s shoutouts. |

        Parameters
        ----------
        user: User
            The user to whom the shoutout will be sent.
        """
        await self._state.http.send_a_shoutout(self._user_id,
                                               self._auth_user_id,
                                               to_broadcaster_id=user.id)

    async def create_clip(self, has_delay: bool = False) -> channels.ClipEdit:
        """
        Create a clip of the broadcaster's stream.

        | Scopes       | Description                 |
        | ------------ | ----------------------------|
        | `clips:edit` | Manage Clips for a channel. |

        Parameters
        ----------
        has_delay: bool
            Whether to create the clip with a delay.

        Returns
        -------
        channels.ClipEdit
            A dictionary representing the clip edit information.
        """
        data: Data[List[channels.ClipEdit]] = await self._state.http.create_clip(self._auth_user_id,
                                                                                 self._user_id,
                                                                                 has_delay=has_delay)
        return data['data'][0]

    async def fetch_schedule(self,
                             segment_ids: Optional[List[str]] = None,
                             start_time: Optional[datetime.datetime] = None,
                             *,
                             first: int = 25) -> AsyncGenerator[streams.Schedule, None]:
        """
        Fetch the broadcaster's stream schedule.

        Parameters
        ----------
        segment_ids: Optional[List[str]]
            A list of segment IDs to filter the schedule.
        start_time: Optional[datetime.datetime]
            The start time to filter the schedule.
        first: int
            The number of items to return per request.

        Yields
        ------
        AsyncGenerator[streams.Schedule, None]
            A dictionary representing a schedule segment.
        """
        kwargs: Dict[str, Any] = {
            'segment_ids': segment_ids,
            'start_time': datetime_to_str(start_time),
            'broadcaster_id': self._user_id,
            'first': first
        }
        while True:
            data: PData[List[streams.Schedule]] = await self._state.http.get_channel_stream_schedule(
                self._auth_user_id,
                **kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    async def get_channel_icalendar(self) -> str:
        """
        Retrieve the iCalendar link for the broadcaster's stream schedule.

        Returns
        -------
        str
            A string representing the iCalendar link for the broadcaster's stream schedule.
        """
        data: str = await self._state.http.get_channel_icalendar(self._auth_user_id, self._user_id)
        return data


class BroadcasterStream(Stream):
    """
    Represents a Broadcaster channel stream.
    """

    __slots__ = ()

    def __init__(self, user_id: str, *, state: ConnectionState) -> None:
        super().__init__(user_id, user_id, state=state)

    async def get_key(self) -> str:
        """
        Retrieve the stream key for the broadcaster.

        | Scopes                    | Description                           |
        | ------------------------- | --------------------------------------|
        | `channel:read:stream_key` | View an authorized user’s stream key. |

        Returns
        -------
        str
            A string representing the stream key.
        """
        data: Data[List[streams.StreamKey]] = await self._state.http.get_stream_key(self._user_id)
        return data['data'][0]['stream_key']

    async def start_commercial(self, length: int = 180) -> streams.CommercialStatus:
        """
        Start a commercial break on the broadcaster's stream.

        | Scopes                    | Description                   |
        | ------------------------- | ------------------------------|
        | `channel:edit:commercial` | Run commercials on a channel. |

        Parameters
        ----------
        length: int
            The duration of the commercial in seconds.

        Returns
        -------
        streams.CommercialStatus
            A dictionary representing the status of the commercial break.
        """
        data: Data[List[streams.CommercialStatus]] = await self._state.http.start_commercial(self._user_id, length)
        return data['data'][0]

    async def start_raid(self, user: User) -> streams.RaidInfo:
        """
        Start a raid to another broadcaster's channel.

        | Scopes                 | Description                               |
        | ---------------------- | ------------------------------------------|
        | `channel:manage:raids` | Manage a channel raiding another channel. |

        Parameters
        ----------
        user: User
            The user to whom the raid will be directed.

        Returns
        -------
        streams.RaidInfo
            A dictionary representing the raid information.
        """
        data: Data[List[streams.RaidInfo]] = await self._state.http.start_raid(self._user_id, user.id)
        return data['data'][0]

    async def cancel_raid(self) -> None:
        """
        Cancel an ongoing raid.

        | Scopes                 | Description                               |
        | ---------------------- | ------------------------------------------|
        | `channel:manage:raids` | Manage a channel raiding another channel. |
        """
        await self._state.http.cancel_raid(self._user_id)

    async def get_ad_schedule(self) -> streams.AdSchedule:
        """
        Retrieve the advertisement schedule for the broadcaster.

        | Scopes             | Description                                        |
        | ------------------ | ---------------------------------------------------|
        | `channel:read:ads` | Read the ads schedule and details on your channel. |
        | `channel:manage:ads` | Manage ads schedule on a channel.                |

        Returns
        -------
        streams.AdSchedule
            A dictionary representing the advertisement schedule.
        """
        data: Data[List[streams.AdSchedule]] = await self._state.http.get_ad_schedule(self._user_id)
        return data['data'][0]

    async def snooze_next_ad(self) -> streams.AdSnooze:
        """
        Snooze the next scheduled advertisement.

        | Scopes               | Description                       |
        | -------------------- | ----------------------------------|
        | `channel:manage:ads` | Manage ads schedule on a channel. |

        Returns
        -------
        streams.AdSnooze
            A dictionary representing the snooze status of the next ad.
        """
        data: Data[List[streams.AdSnooze]] = await self._state.http.snooze_next_ad(self._user_id)
        return data['data'][0]

    @overload
    async def update_schedule_vacation(self,
                                       enable: Literal[False],
                                       vacation_start_time: None = None,
                                       vacation_end_time: None = None,
                                       timezone: None = None) -> None: ...

    @overload
    async def update_schedule_vacation(self,
                                       enable: Literal[True],
                                       vacation_start_time: datetime.datetime,
                                       vacation_end_time: datetime.datetime,
                                       timezone: str) -> None: ...

    async def update_schedule_vacation(self,
                                       enable: bool,
                                       vacation_start_time: Optional[datetime.datetime] = None,
                                       vacation_end_time: Optional[datetime.datetime] = None,
                                       timezone: Optional[str] = None) -> None:
        """
        Update the vacation settings for the broadcast schedule.

        | Scopes                    | Description                         |
        | ------------------------- | ------------------------------------|
        | `channel:manage:schedule` | Manage a channel’s stream schedule. |

        Parameters
        ----------
        enable: bool
            Whether to enable or disable vacation mode.
        vacation_start_time: Optional[datetime.datetime]
            The start time of the vacation.
        vacation_end_time: Optional[datetime.datetime]
            The end time of the vacation.
        timezone: Optional[str]
            The [timezone](https://nodatime.org/TimeZones) of the vacation times.
        """
        await self._state.http.update_channel_stream_schedule(self._user_id, enable,
                                                              datetime_to_str(vacation_start_time),
                                                              datetime_to_str(vacation_end_time),
                                                              timezone)

    async def create_schedule_segment(self,
                                      start_time: datetime.datetime,
                                      timezone: str,
                                      duration: int,
                                      is_recurring: Optional[bool] = None,
                                      category_id: Optional[str] = None,
                                      title: Optional[str] = None) -> streams.Schedule:
        """
        Create a new schedule segment for the broadcaster's stream.

        | Scopes                    | Description                         |
        | ------------------------- | ------------------------------------|
        | `channel:manage:schedule` | Manage a channel’s stream schedule. |

        Parameters
        ----------
        start_time: datetime.datetime
            The start time of the schedule segment.
        timezone: str
            The [timezone](https://nodatime.org/TimeZones) of the start time.
        duration: int
            The duration of the schedule segment in minutes.
        is_recurring: Optional[bool]
            Whether the segment is recurring.
        category_id: Optional[str]
            The category ID of the segment.
        title: Optional[str]
            The title of the segment.

        Returns
        -------
        streams.Schedule
            A dictionary representing the created schedule segment.
        """
        data: Data[List[streams.Schedule]] = await self._state.http.create_channel_schedule_segment(
            self._user_id,
            datetime_to_str(start_time),
            timezone,
            duration,
            is_recurring,
            category_id,
            title)
        return data['data'][0]

    async def update_schedule_segment(self,
                                      segment_id: str,
                                      start_time: Optional[datetime.datetime] = None,
                                      duration: Optional[int] = None,
                                      category_id: Optional[str] = None,
                                      title: Optional[str] = None,
                                      is_canceled: Optional[bool] = None,
                                      timezone: Optional[str] = None) -> streams.Schedule:
        """
        Update an existing schedule segment for the broadcaster's stream.

        | Scopes                    | Description                         |
        | ------------------------- | ------------------------------------|
        | `channel:manage:schedule` | Manage a channel’s stream schedule. |

        Parameters
        ----------
        segment_id: str
            The ID of the segment to update.
        start_time: Optional[datetime.datetime]
            The new start time of the segment.
        duration: Optional[int]
            The new duration of the segment in minutes.
        category_id: Optional[str]
            The new category ID of the segment.
        title: Optional[str]
            The new title of the segment.
        is_canceled: Optional[bool]
            Whether the segment is canceled.
        timezone: Optional[str]
            The [timezone](https://nodatime.org/TimeZones) of the segment times.

        Returns
        -------
        streams.Schedule
            A dictionary representing the updated schedule segment.
        """
        data: Data[List[streams.Schedule]] = await self._state.http.update_channel_schedule_segment(
            self._user_id,
            segment_id,
            start_time,
            duration,
            category_id,
            title,
            is_canceled,
            timezone
        )
        return data['data'][0]

    async def delete_schedule_segment(self, segment_id: str) -> None:
        """
        Delete an existing schedule segment from the broadcaster's stream schedule.

        | Scopes                    | Description                         |
        | ------------------------- | ------------------------------------|
        | `channel:manage:schedule` | Manage a channel’s stream schedule. |

        Parameters
        ----------
        segment_id: str
            The ID of the segment to delete.
        """
        await self._state.http.delete_channel_schedule_segment(self._user_id, segment_id)



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT firstED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from urllib.parse import urlparse
from typing import TYPE_CHECKING
from aiohttp import web
import json

if TYPE_CHECKING:
    from typing import Dict, List, Any

import logging
_logger = logging.getLogger(__name__)


class WebSocket:
    """
    Manages WebSocket connections, client subscriptions, and data distribution.
    """

    __slots__ = ('clients',)

    def __init__(self) -> None:
        self.clients: Dict[str, List[web.WebSocketResponse]] = {}

    def add_client(self, ws: web.WebSocketResponse, filter_param: str) -> None:
        """Add a WebSocket client to the list associated with the filter parameter."""
        if filter_param not in self.clients:
            self.clients[filter_param] = []
        self.clients[filter_param].append(ws)
        _logger.debug('Client added with default `%s`. Total clients: %s',
                      filter_param, len(self.clients[filter_param]))

    def remove_client(self, ws: web.WebSocketResponse) -> None:
        """Remove a WebSocket client from the list of clients for its filter parameter."""
        for filter_param, clients_list in self.clients.items():
            if ws in clients_list:
                clients_list.remove(ws)
                if not clients_list:  # Clean up if no clients left for this filter
                    del self.clients[filter_param]
                _logger.debug('Client removed with default `%s`. Total clients: %s',
                              filter_param, len(clients_list))
                return

    async def handle_connection(self, request: web.Request) -> web.WebSocketResponse:
        """Handle WebSocket connection, prepare, and start interaction."""
        ws = web.WebSocketResponse()
        await ws.prepare(request)

        # Extract the filter parameter from the URL query
        filter_param = request.query.get('default', 'all')
        self.add_client(ws, filter_param)

        try:
            await self.receive_and_echo(ws)
        finally:
            self.remove_client(ws)
            await ws.close()
        return ws

    @staticmethod
    def _format_attachment_url(path: str) -> str:
        """Prepend '/upload/' to local file names; leave URLs unchanged."""
        if urlparse(path).scheme:
            return path
        return f"/upload/{path}"

    async def send_data(self, data: Dict[str, Any], filter_key: str):
        """Send data to all clients associated with the given filter parameter."""
        clients_list = self.clients.get(filter_key, [])

        # Update 'image' and 'audio' fields with formatted URLs if necessary
        for key in ['image', 'audio']:
            if data.get(key):
                data[key] = self._format_attachment_url(data[key])

        for client in clients_list[:]:
            try:
                await client.send_str(json.dumps(data))
            except Exception as exc:
                clients_list.remove(client)
                _logger.exception('Error sending message to WebSocket client', exc)

    @staticmethod
    async def receive_and_echo(ws: web.WebSocketResponse) -> None:
        """Receive and echo back any message received from the WebSocket."""
        try:
            async for msg in ws:
                if msg.type == web.WSMsgType.TEXT:  # type: ignore
                    await ws.send_str(f'Echo: {msg.data}')  # type: ignore
                elif msg.type == web.WSMsgType.ERROR:  # type: ignore
                    _logger.error('WebSocket error: {ws.exception()}')
        except Exception as exc:
            _logger.exception('Exception while receiving WebSocket message', exc)



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from .errors import (HTTPException, TwitchServerError, Forbidden, NotFound, AuthFailure, UnregisteredUser)
from urllib.parse import quote as _uriquote
from . import __version__, __github__
from typing import TYPE_CHECKING
from .utils import json_or_text
import aiohttp
import asyncio
import socket
import time

if TYPE_CHECKING:
    from .types import (Data, DTData, Edata, PData, TData, TPData, activity, analytics, bits, chat, channels,
                        interaction, moderation, search, streams, users, PEdata, TTMData)
    from typing import Any, ClassVar, Coroutine, Dict, List, Literal, Optional, TypeVar, Union

    T = TypeVar('T')
    Response = Coroutine[Any, Any, T]

import logging
_logger = logging.getLogger(__name__)

__all__ = ('HTTPClient',)


class Route:
    """Represents HTTP route."""
    BASE: ClassVar[str] = 'https://api.twitch.tv/helix/'
    OAUTH2: ClassVar[str] = 'https://id.twitch.tv/oauth2/'

    __slots__ = ('method', 'url', 'auth_user_id')

    def __init__(self,
                 auth_user_id: Optional[str],
                 method: str,
                 path: str,
                 oauth2: bool = False,
                 **parameters: Any) -> None:
        self.auth_user_id: Optional[str] = auth_user_id
        self.method: str = method
        base_url = self.OAUTH2 if oauth2 else self.BASE

        if parameters:
            # Filter out None values from parameters
            filtered_params = {k: v for k, v in parameters.items() if v is not None}
            if filtered_params:
                query_string = '?' + '&'.join(
                    f'{k}={_uriquote(str(i))}' for k, v in filtered_params.items() for i in
                    (v if isinstance(v, list) else [v])
                )
            else:
                query_string = ''
        else:
            query_string = ''

        self.url: str = f'{base_url}{path}{query_string}'

    def __repr__(self) -> str:
        return f'<Route method={self.method} url={self.url}>'


class HTTPClient:
    """Represents an asynchronous HTTP client for sending HTTP requests."""

    # Time interval (in seconds) before token validation. -300 seconds by default.
    TOKEN_VALIDATE: ClassVar[int] = 3600

    # Sleep for a reasonable amount of time before checking again
    KEEP_ALIVE_LOOP: ClassVar[int] = 300

    def __init__(self,
                 client_id: str,
                 client_secret: Optional[str],
                 *,
                 loop: Optional[asyncio.AbstractEventLoop] = None,
                 proxy: Optional[str] = None,
                 proxy_auth: Optional[aiohttp.BasicAuth] = None,
                 cli: bool = False,
                 cli_port: int = 8080) -> None:
        # Client settings
        self.client_id: str = client_id
        self.client_secret: Optional[str] = client_secret
        self.user_agent: str = f'twitch.py/{__version__} (GitHub: {__github__})'

        # Debugging and proxy settings
        self.cli: bool = cli
        self.cli_port: int = cli_port
        self.proxy: Optional[str] = proxy
        self.proxy_auth: Optional[aiohttp.BasicAuth] = proxy_auth

        # HTTP session and task management
        self.__session: Optional[aiohttp.ClientSession] = None
        self.__token_keep_alive_task: Optional[asyncio.Task] = None
        self.loop: asyncio.AbstractEventLoop = loop

        # Token storage
        self.__tokens: Dict[str, Dict[str, Any]] = {}

    def get_token(self, user_id) -> Optional[Dict[str, Dict[str, Any]]]:
        """Retrieve token data for a given user."""
        return self.__tokens.get(user_id)

    def add_token(self,
                  user_id, /,
                  access_token: Optional[str],
                  expire_in: Optional[int],
                  refresh_token: Optional[str]) -> None:
        """Add or update token data for a given user."""
        self.__tokens[user_id] = {'access_token': access_token,
                                  'refresh_token': refresh_token,
                                  'expire_in': expire_in,
                                  'validate_in': self.TOKEN_VALIDATE}

    def remove_token(self, user_id) -> None:
        """Remove token data for a given user."""
        if self.__tokens.get(user_id):
            self.__tokens.pop(user_id)

    @property
    def is_open(self) -> bool:
        """Checks if the HTTP session is open. """
        return self.__session is not None and not self.__session.closed

    async def close(self) -> None:
        """Close the HTTP session if open."""
        if self.is_open:
            await self.__session.close()

    async def clear(self) -> None:
        """Clear session, keep-alive task, and token data."""
        if self.__session and self.__session.closed:
            self.__session: Optional[aiohttp.ClientSession] = None

        if self.__token_keep_alive_task is not None and not self.__token_keep_alive_task.done():
            self.__token_keep_alive_task.cancel()

        self.__token_keep_alive_task: Optional[asyncio.Task] = None
        self.__tokens: Dict[str, Dict[str, Any]] = {}

    async def ws_connect(self,
                         url: str,
                         *,
                         compress: int = 0,
                         resume: bool = False) -> aiohttp.ClientWebSocketResponse:
        """Establishes a WebSocket connection to the specified URL with optional compression."""
        kwargs = {
            'max_msg_size': 0,
            'timeout': 30.0,
            'proxy_auth': self.proxy_auth,
            'proxy': self.proxy,
            'autoclose': False,
            'headers': {
                'User-Agent': self.user_agent,
            },
            'compress': compress,
        }
        url: str = f'ws://localhost:{self.cli_port}/ws' if (self.cli and not resume) else url
        return await self.__session.ws_connect(url, **kwargs)

    async def request(self, route: Route, **kwargs: Any) -> Any:
        """Make an HTTP request with the provided route and keyword arguments."""
        method = route.method
        url = route.url

        if 'headers' not in kwargs:
            headers: Dict[str, str] = {'Client-ID': self.client_id}

            if route.auth_user_id:
                token_data = self.__tokens.get(route.auth_user_id)
                if token_data:
                    headers['Authorization'] = f'Bearer {token_data["access_token"]}'
                else:
                    raise UnregisteredUser(
                        'Access token is missing for user %s. Please register the user using `register_user`.'
                        % route.auth_user_id
                    )

            if 'json' in kwargs:
                headers['Content-Type'] = 'application/json'
                kwargs['json'] = kwargs.pop('json')
            kwargs['headers'] = headers

        # Add User-Agent header for the request
        kwargs['headers']['User-Agent'] = self.user_agent

        # Configure proxy settings if provided
        if self.proxy is not None:
            kwargs['proxy'] = self.proxy
        if self.proxy_auth is not None:
            kwargs['proxy_auth'] = self.proxy_auth

        for attempt in range(3):
            try:
                async with self.__session.request(method, url, **kwargs) as response:
                    _logger.debug('%s >> %s with %s has returned status code %s', method, url,
                                  kwargs.get('data'), response.status)

                    data = await json_or_text(response)
                    if 300 > response.status >= 200:
                        _logger.debug('%s << %s has received %s', method, url, data)
                        return data
                    if response.status == 403:
                        raise Forbidden(response, data)
                    elif response.status == 404:
                        raise NotFound(response, data)
                    elif response.status >= 500:
                        raise TwitchServerError(response, data)
                    else:
                        raise HTTPException(response, data)
            except OSError as e:
                if attempt < 2 and e.errno in (54, 10054):
                    await asyncio.sleep(1 + attempt * 2)
                    continue
                raise

    async def initialize_authorization(self,
                                       access_token: Optional[str],
                                       refresh_token: Optional[str]) -> users.OAuthToken:
        """Initialize authorization with the provided access token and refresh token, and manage tokens."""
        if not self.is_open:
            connector = aiohttp.TCPConnector(limit=0, family=socket.AF_INET)
            self.__session = aiohttp.ClientSession(connector=connector)
            _logger.debug('New session has been created.')

        if access_token is None and refresh_token is None:
            raise AuthFailure('Both access token and refresh token are missing. '
                              'Provide at least one token for authorization.')
        while True:
            try:
                # Generate a new token
                if access_token is None and refresh_token:
                    raise AuthFailure

                # Validate the provided access token.
                data: users.OAuthToken = await self.validate_token(access_token)
                # Update the existing token information.
                self.add_token(data['user_id'], access_token, data['expires_in'], refresh_token)
                # Keep the access tokens fresh.
                if self.__token_keep_alive_task is None or self.__token_keep_alive_task.done():
                    self.__token_keep_alive_task = self.loop.create_task(self.token_keep_alive(),
                                                                         name='twitch:http:token_keep_alive')
                    _logger.debug('Keep-alive task has been created.')
                return data

            except (HTTPException, AuthFailure) as exc:
                if isinstance(exc, AuthFailure) or exc.status == 401:
                    if refresh_token and self.client_secret:
                        # Try to refresh the access token.
                        data: users.OAuthRefreshToken = await self.refresh_token(refresh_token)
                        access_token = data['access_token']
                        refresh_token = data['refresh_token']
                        continue
                    else:
                        raise AuthFailure('Improper access token has been passed.') from exc
                raise exc

    def validate_token(self, access_token: str) -> Response[users.OAuthToken]:
        """Validate the provided access token."""
        headers: Dict[str, str] = {'Client-ID': self.client_id, 'Authorization': 'Bearer ' + access_token}
        return self.request(Route(None, 'GET', 'validate', oauth2=True), headers=headers)

    def refresh_token(self, refresh_token: str) -> Response[users.OAuthRefreshToken]:
        """Regenerate the user's access token using the provided refresh token."""
        body: Dict[str, str] = {'grant_type': 'refresh_token',
                                'refresh_token': refresh_token,
                                'client_secret': self.client_secret,
                                'client_id': self.client_id}
        return self.request(Route(None, 'POST', 'token', oauth2=True), data=body)

    async def token_keep_alive(self) -> None:
        """Keeps the tokens alive by regenerating or revalidating when necessary."""

        current_time = time.time()

        while True:
            try:
                for user_id, token_data in self.__tokens.items():
                    elapsed_time = int(time.time() - current_time)
                    self.__tokens[user_id]['validate_in'] = token_data['validate_in'] - elapsed_time
                    self.__tokens[user_id]['expire_in'] = token_data['expire_in'] - elapsed_time

                    # If token is about to expire (within 5 minutes)
                    if self.__tokens[user_id]['expire_in'] <= 300:
                        if self.client_secret and token_data['refresh_token']:
                            try:
                                # Regenerate the token
                                data: users.OAuthRefreshToken = await self.refresh_token(token_data['refresh_token'])
                                self.__tokens[user_id].update({
                                    'access_token': data['access_token'],
                                    'refresh_token': data['refresh_token'],
                                    'expire_in': data['expires_in'],
                                    'validate_in': self.TOKEN_VALIDATE
                                })
                                _logger.debug('Refreshed token for user %s.', user_id)
                            except HTTPException as exc:
                                _logger.warning('Failed to refresh token for user %s: %s', user_id, exc.text)
                                # Invalidate refresh token if regeneration fails
                                self.__tokens[user_id]['refresh_token'] = None
                        else:
                            _logger.warning('Token for user %s is expiring soon due to missing client '
                                            'secret or refresh token.', user_id)

                    if self.__tokens[user_id]['validate_in'] <= 300:
                        try:
                            data: users.OAuthToken = await self.validate_token(self.__tokens[user_id]['access_token'])
                            self.__tokens[user_id].update({
                                'expire_in': data['expires_in'],
                                'validate_in': self.TOKEN_VALIDATE
                            })
                            _logger.debug('Revalidated token for user %s.', user_id)
                        except HTTPException as exc:
                            _logger.warning('Failed to revalidate token for user %s: %s', user_id, exc.text)

            except (OSError, Exception) as exc:
                _logger.exception('An error occurred during the token:keep-alive loop: %s.'
                                  ' Retrying in 30 seconds.', exc)
                # Reset timer.
                current_time = time.time()
                await asyncio.sleep(30)
                continue

            # Reset timer.
            current_time = time.time()
            await asyncio.sleep(self.KEEP_ALIVE_LOOP)

    @staticmethod
    def get_subscription_info(event: str) -> Optional[Dict[str, Any]]:
        # Warning: This dictionary may be updated anytime based on new event types or API changes.
        # It maps subscription types to their respective Twitch event name and version.
        subscriptions: Dict[str, Dict[str, Any]] = {
            'automod_message_hold': {
                'name': 'automod.message.hold',
                'version': '1',
                'condition': {'broadcaster': 'moderator_user_id', 'user': 'broadcaster_user_id'}
            },
            'automod_message_update': {
                'name': 'automod.message.update',
                'version': '1',
                'condition': {'broadcaster': 'moderator_user_id', 'user': 'broadcaster_user_id'}
            },
            'automod_settings_update': {
                'name': 'automod.settings.update',
                'version': '1',
                'condition': {'broadcaster': 'moderator_user_id', 'user': 'broadcaster_user_id'}
            },
            'automod_terms_update': {
                'name': 'automod.terms.update',
                'version': '1',
                'condition': {'broadcaster': 'moderator_user_id', 'user': 'broadcaster_user_id'}
            },
            'channel_update': {
                'name': 'channel.update',
                'version': '2',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'follow': {
                'name': 'channel.follow',
                'version': '2',
                'condition': {'broadcaster': 'moderator_user_id', 'user': 'broadcaster_user_id'}
            },
            'ad_break_begin': {
                'name': 'channel.ad_break.begin',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'chat_clear': {
                'name': 'channel.chat.clear',
                'version': '1',
                'condition': {'broadcaster': 'user_id', 'user': 'broadcaster_user_id'}
            },
            'chat_clear_user_messages': {
                'name': 'channel.chat.clear_user_messages',
                'version': '1',
                'condition': {'broadcaster': 'user_id', 'user': 'broadcaster_user_id'}
            },
            'chat_message': {
                'name': 'channel.chat.message',
                'version': '1',
                'condition': {'broadcaster': 'user_id', 'user': 'broadcaster_user_id'}
            },
            'chat_message_delete': {
                'name': 'channel.chat.message_delete',
                'version': '1',
                'condition': {'broadcaster': 'user_id', 'user': 'broadcaster_user_id'}
            },
            'chat_notification': {
                'name': 'channel.chat.notification',
                'version': '1',
                'condition': {'broadcaster': 'user_id', 'user': 'broadcaster_user_id'}
            },
            'chat_settings_update': {
                'name': 'channel.chat_settings.update',
                'version': '1',
                'condition': {'broadcaster': 'user_id', 'user': 'broadcaster_user_id'}
            },
            'chat_user_message_hold': {
                'name': 'channel.chat.user_message_hold',
                'version': '1',
                'condition': {'broadcaster': 'user_id', 'user': 'broadcaster_user_id'}
            },
            'chat_user_message_update': {
                'name': 'channel.chat.user_message_update',
                'version': '1',
                'condition': {'broadcaster': 'user_id', 'user': 'broadcaster_user_id'}
            },
            'subscribe': {
                'name': 'channel.subscribe',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'subscription_end': {
                'name': 'channel.subscription.end',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'subscription_gift': {
                'name': 'channel.subscription.gift',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'subscription_message': {
                'name': 'channel.subscription.message',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'cheer': {
                'name': 'channel.cheer',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'raid': {
                'name': 'channel.raid',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'to_broadcaster_user_id'}
            },
            'ban': {
                'name': 'channel.ban',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'unban': {
                'name': 'channel.unban',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'unban_request_create': {
                'name': 'channel.unban_request.create',
                'version': '1',
                'condition': {'broadcaster': 'moderator_user_id', 'user': 'broadcaster_user_id'}
            },
            'unban_request_resolve': {
                'name': 'channel.unban_request.resolve',
                'version': '1',
                'condition': {'broadcaster': 'moderator_user_id', 'user': 'broadcaster_user_id'}
            },
            'moderate': {
                'name': 'channel.moderate',
                'version': '2',
                'condition': {'broadcaster': 'moderator_user_id', 'user': 'broadcaster_user_id'}
            },
            'moderator_add': {
                'name': 'channel.moderator.add',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'moderator_remove': {
                'name': 'channel.moderator.remove',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'points_automatic_reward_redemption_add': {
                'name': 'channel.channel_points_automatic_reward_redemption.add',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'points_reward_add': {
                'name': 'channel.channel_points_custom_reward.add',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'points_reward_update': {
                'name': 'channel.channel_points_custom_reward.update',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'points_reward_remove': {
                'name': 'channel.channel_points_custom_reward.remove',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'points_reward_redemption_add': {
                'name': 'channel.channel_points_custom_reward_redemption.add',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'points_reward_redemption_update': {
                'name': 'channel.channel_points_custom_reward_redemption.update',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'poll_begin': {
                'name': 'channel.poll.begin',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'poll_progress': {
                'name': 'channel.poll.progress',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'poll_end': {
                'name': 'channel.poll.end',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'prediction_begin': {
                'name': 'channel.prediction.begin',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'prediction_progress': {
                'name': 'channel.prediction.progress',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'prediction_lock': {
                'name': 'channel.prediction.lock',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'prediction_end': {
                'name': 'channel.prediction.end',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'suspicious_user_message': {
                'name': 'channel.suspicious_user.message',
                'version': '1',
                'condition': {'broadcaster': 'moderator_user_id', 'user': 'broadcaster_user_id'}
            },
            'suspicious_user_update': {
                'name': 'channel.suspicious_user.update',
                'version': '1',
                'condition': {'broadcaster': 'moderator_user_id', 'user': 'broadcaster_user_id'}
            },
            'vip_add': {
                'name': 'channel.vip.add',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'vip_remove': {
                'name': 'channel.vip.remove',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'warning_acknowledge': {
                'name': 'channel.warning.acknowledge',
                'version': '1',
                'condition': {'broadcaster': 'moderator_user_id', 'user': 'broadcaster_user_id'}
            },
            'warning_send': {
                'name': 'channel.warning.send',
                'version': '1',
                'condition': {'broadcaster': 'moderator_user_id', 'user': 'broadcaster_user_id'}
            },
            'charity_campaign_donate': {
                'name': 'channel.charity_campaign.donate',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'charity_campaign_start': {
                'name': 'channel.charity_campaign.start',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'charity_campaign_progress': {
                'name': 'channel.charity_campaign.progress',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'charity_campaign_stop': {
                'name': 'channel.charity_campaign.stop',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'goal_begin': {
                'name': 'channel.goal.begin',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'goal_progress': {
                'name': 'channel.goal.progress',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'goal_end': {
                'name': 'channel.goal.end',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'hype_train_begin': {
                'name': 'channel.hype_train.begin',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'hype_train_progress': {
                'name': 'channel.hype_train.progress',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'hype_train_end': {
                'name': 'channel.hype_train.end',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'shield_mode_begin': {
                'name': 'channel.shield_mode.begin',
                'version': '1',
                'condition': {'broadcaster': 'moderator_user_id', 'user': 'broadcaster_user_id'}
            },
            'shield_mode_end': {
                'name': 'channel.shield_mode.end',
                'version': '1',
                'condition': {'broadcaster': 'moderator_user_id', 'user': 'broadcaster_user_id'}
            },
            'shoutout_create': {
                'name': 'channel.shoutout.create',
                'version': '1',
                'condition': {'broadcaster': 'moderator_user_id', 'user': 'broadcaster_user_id'}
            },
            'shoutout_received': {
                'name': 'channel.shoutout.receive',
                'version': '1',
                'condition': {'broadcaster': 'moderator_user_id', 'user': 'broadcaster_user_id'}
            },
            'stream_online': {
                'name': 'stream.online',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'stream_offline': {
                'name': 'stream.offline',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'broadcaster_user_id'}
            },
            'user_authorization_grant': {
                'name': 'user.authorization.grant',
                'version': '1',
                'condition': {'broadcaster': 'broadcaster_id', 'user': None}
            },
            'user_authorization_revoke': {
                'name': 'user.authorization.revoke',
                'version': '1',
                'condition': {'broadcaster': 'broadcaster_id', 'user': None}
            },
            'user_update': {
                'name': 'user.update',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'user_id'}
            },
            'whisper_received': {
                'name': 'user.whisper.message',
                'version': '1',
                'condition': {'broadcaster': None, 'user': 'user_id'}
            }
        }
        return subscriptions.get(event)

    def create_subscription(
            self,
            __id: str,
            broadcaster_id: str,
            user_id: str,
            session_id: str,
            *,
            subscription_type: str,
            subscription_version: str,
            subscription_condition: Dict[str, Any]
    ) -> Response[TTMData[List[users.EventSubSubscription]]]:
        """Create an EventSub Websocket Subscription."""
        route = Route(__id, 'POST', 'eventsub/subscriptions')
        if self.cli:
            route.url = f'http://localhost:{self.cli_port}/eventsub/subscriptions'

        condition = {}

        # Ensure 'broadcaster' key is properly assigned
        client_key = subscription_condition.get('broadcaster')
        if client_key:
            condition[client_key] = broadcaster_id

        # Ensure 'user' key is properly assigned
        user_key = subscription_condition.get('user')
        if user_key:
            condition[user_key] = user_id

        body = {
            'type': subscription_type,
            'version': subscription_version,
            'condition': condition,
            'transport': {
                'method': 'websocket',
                'session_id': session_id
            }
        }
        return self.request(route=route, json=body)

    def delete_subscription(self, __id: str, subscription_id: str) -> Response[None]:
        """Delete an EventSub Websocket Subscription."""
        params: Dict[str, Any] = {
            'id': subscription_id
        }
        return self.request(Route(__id, 'DELETE', 'eventsub/subscriptions', **params))

    # ---------------------------------------
    #               GENERAL API
    # ---------------------------------------

    # Ads
    def start_commercial(self,
                         broadcaster_id: str,
                         length: int = 180
                         ) -> Response[Data[List[streams.CommercialStatus]]]:
        body = {
            'broadcaster_id': broadcaster_id,
            'length': length
        }
        return self.request(Route(broadcaster_id, 'POST', 'channels/commercial'), data=body)

    def get_ad_schedule(self, broadcaster_id: str) -> Response[Data[List[streams.AdSchedule]]]:
        return self.request(Route(broadcaster_id, 'GET', 'channels/ads', broadcaster_id=broadcaster_id))

    def snooze_next_ad(self, broadcaster_id: str) -> Response[Data[List[streams.AdSnooze]]]:
        return self.request(Route(broadcaster_id, 'POST', 'channels/ads/schedule/snooze',
                                  broadcaster_id=broadcaster_id))

    # Analytics
    def get_extension_analytics(self,
                                __id: str,
                                extension_id: Optional[str] = None,
                                analytics_type: Literal['overview_v2'] = 'overview_v2',
                                started_at: Optional[str] = None,
                                ended_at: Optional[str] = None,
                                first: int = 20,
                                after: Optional[str] = None) -> Response[PData[List[analytics.Extension]]]:
        params: Dict[str, Any] = {
            'extension_id': extension_id,
            'type': analytics_type,
            'started_at': started_at,
            'ended_at': ended_at,
            'first': first,
            'after': after
        }
        return self.request(Route(__id, 'GET', 'analytics/extensions', **params))

    def get_game_analytics(self, __id: str,
                           game_id: Optional[str] = None,
                           analytics_type: Literal['overview_v2'] = 'overview_v2',
                           started_at: Optional[str] = None,
                           ended_at: Optional[str] = None,
                           first: int = 20,
                           after: Optional[str] = None) -> Response[PData[List[analytics.Game]]]:
        params: Dict[str, Any] = {
            'game_id': game_id,
            'type': analytics_type,
            'started_at': started_at,
            'ended_at': ended_at,
            'first': first,
            'after': after
        }
        return self.request(Route(__id, 'GET', 'analytics/games', **params))

    # Bits
    def get_bits_leaderboard(self, __id: str,
                             period: Optional[Literal['day', 'week', 'month', 'year', 'all']],
                             started_at: Optional[str],
                             user_id: Optional[str],
                             count: int = 10) -> Response[DTData[List[bits.Leaderboard]]]:
        params: Dict[str, Any] = {'count': count,
                                  'period': period,
                                  'started_at': started_at,
                                  'user_id': user_id
                                  }
        return self.request(Route(__id, 'GET', 'bits/leaderboard', **params))

    def get_cheermotes(self, __id: str, broadcaster_id: Optional[str] = None) -> Response[Data[List[bits.Cheermote]]]:
        return self.request(Route(__id, 'GET', 'bits/cheermotes', broadcaster_id=broadcaster_id))

    # Channel
    def get_channel_information(self, __id: str, broadcaster_ids: List[str]
                                ) -> Response[Data[List[channels.ChannelInfo]]]:
        return self.request(Route(__id, 'GET', 'channels', broadcaster_id=broadcaster_ids))

    def modify_channel_information(self,
                                   broadcaster_id: str,
                                   category_id: Optional[str] = None,
                                   broadcaster_language: Optional[str] = None,
                                   title: Optional[str] = None,
                                   delay: Optional[int] = None,
                                   tags: Optional[List[str]] = None,
                                   content_classification_labels: Optional[List[channels.CCL]] = None,
                                   is_branded_content: Optional[bool] = None) -> Response[None]:
        body = {
            'game_id': category_id,
            'broadcaster_language': broadcaster_language,
            'title': title,
            'delay': delay,
            'tags': tags,
            'content_classification_labels': content_classification_labels,
            is_branded_content: is_branded_content,
        }
        body = {key: value for key, value in body.items() if value is not None}
        return self.request(Route(broadcaster_id, 'PATCH', 'channels', broadcaster_id=broadcaster_id),
                            data=body)

    def get_channel_editors(self, broadcaster_id: str) -> Response[Data[List[channels.Editor]]]:
        return self.request(Route(broadcaster_id, 'GET', 'channels/editors',
                                                  broadcaster_id=broadcaster_id))

    def get_followed_channels(self,
                              user_id: str,
                              *,
                              broadcaster_id: Optional[str] = None,
                              after: Optional[str] = None,
                              first: int = 20) -> Response[TData[List[channels.Follows]]]:
        params: Dict[str, Any] = {'user_id': user_id,
                                  'broadcaster_id': broadcaster_id,
                                  'after': after,
                                  'first': first}

        return self.request(Route(user_id, 'GET', 'channels/followed', **params))

    def get_channel_followers(self, __id: str,
                              broadcaster_id: str,
                              *,
                              user_id: Optional[str] = None,
                              after: Optional[str] = None,
                              first: int = 20) -> Response[TData[List[channels.Follower]]]:
        params: Dict[str, Any] = {'broadcaster_id': broadcaster_id,
                                  'user_id': user_id,
                                  'after': after,
                                  'first': first}

        return self.request(Route(__id, 'GET', 'channels/followers', **params))

    # Channel Points
    def create_custom_rewards(self,
                              broadcaster_id: str,
                              title: str,
                              cost: int,
                              is_enabled: bool = True,
                              background_color: Optional[str] = None,
                              is_user_input_required: bool = False,
                              prompt: Optional[str] = None,
                              is_max_per_stream_enabled: bool = False,
                              max_per_stream: Optional[int] = None,
                              is_max_per_user_per_stream_enabled: bool = False,
                              max_per_user_per_stream: Optional[int] = None,
                              is_global_cooldown_enabled: bool = False,
                              global_cooldown_seconds: Optional[int] = None,
                              should_redemptions_skip_request_queue: bool = False
                              ) -> Response[Data[List[interaction.Reward]]]:
        body: Dict[str, Any] = {
            'title': title,
            'cost': cost,
            'prompt': prompt,
            'is_enabled': is_enabled,
            'background_color': background_color,
            'is_user_input_required': is_user_input_required,
            'is_max_per_stream_enabled': is_max_per_stream_enabled,
            'max_per_stream': max_per_stream,
            'is_max_per_user_per_stream_enabled': is_max_per_user_per_stream_enabled,
            'max_per_user_per_stream': max_per_user_per_stream,
            'is_global_cooldown_enabled': is_global_cooldown_enabled,
            'global_cooldown_seconds': global_cooldown_seconds,
            'should_redemptions_skip_request_queue': should_redemptions_skip_request_queue
        }
        body = {key: value for key, value in body.items() if value is not None}
        return self.request(Route(broadcaster_id, 'POST', 'channel_points/custom_rewards',
                                  broadcaster_id=broadcaster_id), data=body)

    def delete_custom_reward(self, broadcaster_id: str, reward_id: str) -> Response[None]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'id': reward_id
        }
        return self.request(Route(broadcaster_id, 'DELETE', 'channel_points/custom_rewards', **params))

    def get_custom_reward(self,
                          broadcaster_id: str,
                          reward_ids: Optional[List[str]],
                          only_manageable_rewards: bool = False
                          ) -> Response[Data[List[interaction.Reward]]]:

        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'id': reward_ids,
            'only_manageable_rewards': only_manageable_rewards
        }

        return self.request(Route(broadcaster_id, 'GET', 'channel_points/custom_rewards', **params))

    def get_custom_reward_redemption(self,
                                     broadcaster_id: str,
                                     reward_id: str,
                                     redemption_ids: Optional[List[str]],
                                     status: Optional[Literal['canceled', 'fulfilled', 'unfulfilled']] = None,
                                     after: Optional[str] = None,
                                     sort: Literal['oldest', 'newest'] = 'oldest',
                                     first: int = 20
                                     ) -> Response[PData[List[interaction.RewardRedemption]]]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'reward_id': reward_id,
            'status': status.upper() if status is not None else None,
            'redemption_ids': redemption_ids,
            'after': after,
            'sort': sort.upper(),
            'first': first,
        }

        return self.request(Route(broadcaster_id, 'GET', 'channel_points/custom_rewards/redemptions',
                                                  **params))

    def update_custom_reward(self,
                             broadcaster_id: str,
                             reward_id: str,
                             title: Optional[str] = None,
                             cost: Optional[int] = None,
                             is_enabled: Optional[bool] = None,
                             background_color: Optional[str] = None,
                             is_user_input_required: Optional[bool] = None,
                             prompt: Optional[str] = None,
                             is_max_per_stream_enabled: Optional[bool] = None,
                             max_per_stream: Optional[int] = None,
                             is_max_per_user_per_stream_enabled: Optional[bool] = None,
                             max_per_user_per_stream: Optional[int] = None,
                             is_global_cooldown_enabled: Optional[bool] = None,
                             global_cooldown_seconds: Optional[int] = None,
                             should_redemptions_skip_request_queue: Optional[bool] = None
                             ) -> Response[Data[List[interaction.Reward]]]:

        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'id': reward_id,
        }

        body: Dict[str, Any] = {
            'title': title,
            'cost': cost,
            'prompt': prompt,
            'is_enabled': is_enabled,
            'background_color': background_color,
            'is_user_input_required': is_user_input_required,
            'is_max_per_stream_enabled': is_max_per_stream_enabled,
            'max_per_stream': max_per_stream,
            'is_max_per_user_per_stream_enabled': is_max_per_user_per_stream_enabled,
            'max_per_user_per_stream': max_per_user_per_stream,
            'is_global_cooldown_enabled': is_global_cooldown_enabled,
            'global_cooldown_seconds': global_cooldown_seconds,
            'should_redemptions_skip_request_queue': should_redemptions_skip_request_queue
        }
        body = {key: value for key, value in body.items() if value is not None}
        return self.request(Route(broadcaster_id, 'PATCH', 'channel_points/custom_rewards', **params),
                            data=body)

    def update_redemption_status(self,
                                 broadcaster_id: str,
                                 reward_id: str,
                                 redemption_ids: List[str],
                                 *,
                                 status: Literal['canceled', 'fulfilled']
                                 ) -> Response[Data[List[interaction.RewardRedemption]]]:

        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'id': redemption_ids,
            'reward_id': reward_id
        }

        body: Dict[str, Any] = {
            'status': status.upper(),
        }
        return self.request(Route(broadcaster_id, 'PATCH', 'channel_points/custom_rewards/redemptions',
                                                  **params),
                            data=body)

    # Charity
    def get_charity_campaign(self, broadcaster_id: str) -> Response[Data[List[activity.Charity]]]:
        return self.request(Route(broadcaster_id, 'GET', 'charity/campaigns',
                                                  broadcaster_id=broadcaster_id))

    def get_charity_campaign_donations(self,
                                       broadcaster_id: str,
                                       after: Optional[str],
                                       first: int = 20) -> Response[PData[List[activity.CharityDonation]]]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'first': first,
            'after': after
        }

        return self.request(Route(broadcaster_id, 'GET', 'charity/donations', **params))

    # Chat
    def get_chatters(self,
                     broadcaster_id: str,
                     moderator_id: str,
                     after: Optional[str] = None,
                     first: int = 20) -> Response[TData[List[users.SpecificUser]]]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'moderator_id': moderator_id,
            'first': first,
            'after': after
        }
        return self.request(Route(moderator_id, 'GET', 'chat/chatters', **params))

    def get_channel_emotes(self, __id: str, broadcaster_id: str) -> Response[Edata[List[chat.Emote]]]:
        return self.request(Route(__id, 'GET', 'chat/emotes', broadcaster_id=broadcaster_id))

    def get_global_emotes(self, __id: str) -> Response[Edata[List[chat.Emote]]]:
        return self.request(Route(__id, 'GET', 'chat/emotes/global'))

    def get_emote_sets(self, __id: str, emote_set_ids: List[str]) -> Response[Edata[List[chat.Emote]]]:
        return self.request(Route(__id, 'GET', 'chat/emotes/set', emote_set_id=emote_set_ids))

    def get_user_emotes(self,
                        user_id: str,
                        broadcaster_id: Optional[str] = None,
                        after: Optional[str] = None) -> Response[PEdata[List[chat.Emote]]]:
        params: Dict[str, Any] = {
            'user_id': user_id,
            'broadcaster_id': broadcaster_id,
            'after': after
        }
        return self.request(Route(user_id, 'GET', 'chat/emotes/user', **params))

    def get_channel_chat_badges(self, __id: str, broadcaster_id: str) -> Response[Data[List[chat.Badge]]]:
        return self.request(Route(__id, 'GET', 'chat/badges', broadcaster_id=broadcaster_id))

    def get_global_chat_badges(self, __id: str) -> Response[Data[List[chat.Badge]]]:
        return self.request(Route(__id, 'GET', 'chat/badges/global'))

    def get_chat_settings(self,
                          broadcaster_id: str,
                          moderator_id: Optional[str]) -> Response[Data[List[chat.Settings]]]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'moderator_id': moderator_id
        }

        return self.request(Route(moderator_id, 'GET', 'chat/settings', **params))

    def update_chat_settings(
            self,
            broadcaster_id: str,
            moderator_id: str,
            *,
            emote_mode: Optional[bool] = None,
            follower_mode: Optional[bool] = None,
            follower_mode_duration: Optional[int] = None,
            non_moderator_chat_delay: Optional[bool] = None,
            non_moderator_chat_delay_duration: Optional[int] = None,
            slow_mode: Optional[bool] = None,
            slow_mode_wait_time: Optional[int] = None,
            subscriber_mode: Optional[bool] = None,
            unique_chat_mode: Optional[bool] = None) -> Response[Data[List[chat.Settings]]]:

        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'moderator_id': moderator_id
        }

        body: Dict[str, Any] = {
            'emote_mode': emote_mode,
            'follower_mode': follower_mode,
            'follower_mode_duration': follower_mode_duration if follower_mode else None,
            'non_moderator_chat_delay': non_moderator_chat_delay,
            'non_moderator_chat_delay_duration': (non_moderator_chat_delay_duration
                                                  if non_moderator_chat_delay else None),
            'slow_mode': slow_mode,
            'slow_mode_wait_time': slow_mode_wait_time if slow_mode else None,
            'subscriber_mode': subscriber_mode,
            'unique_chat_mode': unique_chat_mode
        }

        body = {key: value for key, value in body.items() if value is not None}
        return self.request(Route(moderator_id, 'PATCH', 'chat/settings', **params), data=body)

    def send_chat_announcement(self,
                               broadcaster_id: str,
                               moderator_id: str,
                               *,
                               message: str,
                               color: Literal['blue', 'green', 'orange', 'purple', 'primary'] = 'primary'
                               ) -> Response[None]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'moderator_id': moderator_id
        }

        body: Dict[str, Any] = {
            'message': message,
            'color': color,
        }
        return self.request(Route(moderator_id, 'POST', 'chat/announcements', **params), data=body)

    def send_a_shoutout(self,
                        from_broadcaster_id: str,
                        moderator_id: str, *,
                        to_broadcaster_id: str,
                        ) -> Response[None]:
        params: Dict[str, Any] = {
            'from_broadcaster_id': from_broadcaster_id,
            'to_broadcaster_id': to_broadcaster_id,
            'moderator_id': moderator_id
        }
        return self.request(Route(moderator_id, 'POST', 'chat/shoutouts', **params))

    def send_chat_message(self,
                          broadcaster_id: str,
                          sender_id: str,
                          text: str,
                          reply_parent_message_id: Optional[str] = None
                          ) -> Response[Data[List[chat.SendMessageStatus]]]:

        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'sender_id': sender_id,
            'message': text,
            'reply_parent_message_id': reply_parent_message_id
        }

        return self.request(Route(sender_id, 'POST', 'chat/messages', **params))

    def get_user_chat_color(self, __id: str, user_ids: List[str]) -> Response[Data[List[chat.UserChatColor]]]:
        return self.request(Route(__id, 'GET', 'chat/color', user_id=user_ids))

    def update_user_chat_color(self,
                               user_id: str,
                               color: Union[str, chat.UserChatColors]) -> Response[None]:
        return self.request(Route(user_id, 'PUT', 'chat/color', user_id=user_id, color=color))

    # Clips
    def create_clip(self, __id: str,
                    broadcaster_id: str,
                    *,
                    has_delay: bool = False
                    ) -> Response[Data[List[channels.ClipEdit]]]:
        return self.request(Route(__id, 'POST', 'clips', broadcaster_id=broadcaster_id,
                                        has_delay=has_delay))

    def get_clips(self, __id: str,
                  broadcaster_id: Optional[str] = None,
                  game_id: Optional[str] = None,
                  clip_ids: Optional[List[str]] = None,
                  started_at: Optional[str] = None,
                  ended_at: Optional[str] = None,
                  first: int = 20,
                  before: Optional[str] = None,
                  after: Optional[str] = None,
                  is_featured: Optional[bool] = None
                  ) -> Response[PData[List[channels.Clip]]]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'game_id': game_id,
            'ids': clip_ids,
            'started_at': started_at,
            'ended_at': ended_at,
            'first': first,
            'before': before,
            'after': after,
            'is_featured': is_featured
        }
        return self.request(Route(__id, 'GET', 'clips', **params))

    # CCLs
    def get_content_classification_labels(self, __id: str,
                                          locale: streams.Locale = 'en-US'
                                          ) -> Response[Data[List[streams.CCLInfo]]]:
        return self.request(Route(__id, 'GET', 'content_classification_labels', locale=locale))

    # Entitlements
    def get_drops_entitlements(self,
                               entitlement_ids: Optional[List[str]] = None,
                               user_id: Optional[str] = None,
                               game_id: Optional[str] = None,
                               fulfillment_status: Optional[Literal['claimed', 'fulfilled']] = None,
                               after: Optional[str] = None,
                               first: int = 200
                               ) -> Response[PData[List[activity.Entitlement]]]:
        params: Dict[str, Any] = {
            'id': entitlement_ids,
            'user_id': user_id,
            'game_id': game_id,
            'fulfillment_status': fulfillment_status.upper() if fulfillment_status is not None else None,
            'after': after,
            'first': first
        }
        return self.request(Route(user_id, 'GET', 'entitlements/drops', **params))

    def update_drops_entitlements(self, __id: str,
                                  entitlement_ids: List[str],
                                  fulfillment_status: Literal['claimed', 'fulfilled']
                                  ) -> Response[Data[List[activity.EntitlementsUpdate]]]:
        params: Dict[str, Any] = {
            'entitlement_ids': entitlement_ids,
            'fulfillment_status': fulfillment_status.upper(),
        }
        return self.request(Route(__id, 'PATCH', 'entitlements/drops', **params))

    # Games
    def get_top_games(self, __id: str,
                      after: Optional[str] = None,
                      first: int = 20) -> Response[PData[List[search.Game]]]:

        params: Dict[str, Any] = {
            'first': first,
            'after': after,
        }
        return self.request(Route(__id, 'GET', 'games/top', **params))

    def get_games(self, __id: str,
                  game_ids: Optional[List[str]] = None,
                  names: Optional[List[str]] = None,
                  igdb_ids: Optional[List[str]] = None) -> Response[Data[List[search.Game]]]:
        params: Dict[str, Any] = {
            'id': game_ids,
            'name': names,
            'igdb_id': igdb_ids
        }
        return self.request(Route(__id, 'GET', 'games', **params))

    # Goals
    def get_creator_goals(self, broadcaster_id: str) -> Response[Data[List[activity.Goal]]]:
        return self.request(Route(broadcaster_id, 'GET', 'goals', broadcaster_id=broadcaster_id))

    # Hype Train
    def get_hype_train_events(self,
                              broadcaster_id: str,
                              after: Optional[str] = None,
                              first: int = 20) -> Response[PData[List[interaction.HypeTrain]]]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'after': after,
            'first': first
        }
        return self.request(Route(broadcaster_id, 'GET', 'hypetrain/events', **params))

    # Moderation
    def check_automod_status(self,
                             broadcaster_id: str,
                             messages: List[str]
                             ) -> Response[Data[List[moderation.AutoModMessageStatus]]]:
        body = {'data': [{'msg_id': str(abs(hash(msg))), 'msg_text': msg} for msg in messages]}
        return self.request(Route(broadcaster_id, 'POST', 'moderation/enforcements/status',
                                  broadcaster_id=broadcaster_id),
                            json=body)

    def manage_held_automod_messages(self,
                                     user_id: str,
                                     msg_id: str,
                                     action: Literal['allow', 'deny']
                                     ) -> Response[None]:
        params: Dict[str, Any] = {
            'user_id': user_id,
            'msg_id': msg_id,
            'action': action.upper()
        }
        return self.request(Route(user_id, 'POST', 'moderation/automod/message', **params))

    def get_automod_settings(self,
                             broadcaster_id: str,
                             moderator_id: str) -> Response[Data[List[moderation.AutoModSettings]]]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'moderator_id': moderator_id,
        }
        return self.request(Route(moderator_id, 'GET', 'moderation/automod/settings', **params))

    def update_automod_settings(self,
                                broadcaster_id: str,
                                moderator_id: str,
                                *,
                                overall_level: Optional[int] = None,
                                aggression: Optional[int] = None,
                                bullying: Optional[int] = None,
                                disability: Optional[int] = None,
                                misogyny: Optional[int] = None,
                                race_ethnicity_or_religion: Optional[int] = None,
                                sex_based_terms: Optional[int] = None,
                                sexuality_sex_or_gender: Optional[int] = None,
                                swearing: Optional[int] = None
                                ) -> Response[Data[List[moderation.AutoModSettings]]]:

        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'moderator_id': moderator_id,
        }

        if overall_level is None:
            body: Dict[str, Any] = {
                'aggression': aggression,
                'bullying': bullying,
                'disability': disability,
                'misogyny': misogyny,
                'race_ethnicity_or_religion': race_ethnicity_or_religion,
                'sex_based_terms': sex_based_terms,
                'sexuality_sex_or_gender': sexuality_sex_or_gender,
                'swearing': swearing}
            body = {key: value for key, value in body.items() if value is not None}
        else:
            body: Dict[str, Any] = {'overall_level': overall_level}
        return self.request(Route(moderator_id, 'PUT', 'moderation/automod/settings', **params), data=body)

    def get_banned_users(self, __id: str,
                         broadcaster_id: str,
                         *,
                         user_ids: Optional[List[str]] = None,
                         first: int = 20,
                         before: Optional[str] = None,
                         after: Optional[str] = None) -> Response[PData[List[moderation.BannedUser]]]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'user_id': user_ids,
            'first': first,
            'before': before,
            'after': after,
        }

        return self.request(Route(__id, 'GET', 'moderation/banned', **params))

    def ban_user(self,
                 broadcaster_id: str,
                 moderator_id: str,
                 user_id: str,
                 *,
                 duration: Optional[int] = None,
                 reason: Optional[str] = None) -> Response[Data[List[moderation.BanUser]]]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'moderator_id': moderator_id,
        }

        body: Dict[str, Any] = {
            'user_id': user_id,
            'duration': duration,
            'reason': reason
        }
        body = {key: value for key, value in body.items() if value is not None}
        return self.request(Route(moderator_id, 'POST', 'moderation/bans', **params), json={'data': body})

    def unban_user(self,
                   broadcaster_id: str,
                   moderator_id: str,
                   user_id: str) -> Response[None]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'moderator_id': moderator_id,
            'user_id': user_id
        }
        return self.request(Route(moderator_id, 'DELETE', 'moderation/bans', **params))

    def get_unban_requests(self,
                           broadcaster_id: str,
                           moderator_id: str,
                           status: Literal['pending', 'approved', 'denied', 'acknowledged', 'canceled'],
                           *,
                           user_id: Optional[str] = None,
                           after: Optional[str] = None,
                           first: int = 20
                           ) -> Response[PData[List[moderation.UnBanRequest]]]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'moderator_id': moderator_id,
            'status': status,
            'user_id': user_id,
            'after': after,
            'first': first
        }

        return self.request(Route(moderator_id, 'GET', 'moderation/unban_requests', **params))

    def resolve_unban_requests(self,
                               broadcaster_id: str,
                               moderator_id: str,
                               *,
                               unban_request_id: str,
                               status: Literal['approved', 'denied'],
                               resolution_text: Optional[str] = None
                               ) -> Response[Data[List[moderation.UnBanRequest]]]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'moderator_id': moderator_id,
            'unban_request_id': unban_request_id,
            'status': status,
            'resolution_text': resolution_text,
        }

        return self.request(Route(moderator_id, 'PATCH', 'moderation/unban_requests', **params))

    def get_blocked_terms(self,
                          broadcaster_id: str,
                          moderator_id: str,
                          first: int = 20,
                          after: Optional[str] = None
                          ) -> Response[PData[List[moderation.BlockedTerm]]]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'moderator_id': moderator_id,
            'first': first,
            'after': after
        }

        return self.request(Route(moderator_id, 'GET', 'moderation/blocked_terms', **params))

    def add_blocked_term(self,
                         broadcaster_id: str,
                         moderator_id: str,
                         *,
                         text: str
                         ) -> Response[Data[List[moderation.BlockedTerm]]]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'moderator_id': moderator_id
        }

        body: Dict[str, Any] = {
            'text': text
        }
        return self.request(Route(moderator_id, 'POST', 'moderation/blocked_terms', **params), data=body)

    def remove_blocked_term(self,
                            broadcaster_id: str,
                            moderator_id: str,
                            term_id: str
                            ) -> Response[None]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'moderator_id': moderator_id,
            'id': term_id
        }
        return self.request(Route(moderator_id, 'DELETE', 'moderation/blocked_terms', **params))

    def delete_chat_messages(self,
                             broadcaster_id: str,
                             moderator_id: str,
                             *,
                             message_id: Optional[str] = None
                             ) -> Response[None]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'moderator_id': moderator_id,
            'message_id': message_id
        }

        return self.request(Route(moderator_id, 'DELETE', 'moderation/chat', **params))

    def get_moderated_channels(self,
                               user_id: str,
                               after: Optional[str] = None,
                               first: int = 20
                               ) -> Response[PData[List[users.Broadcaster]]]:
        params: Dict[str, Any] = {
            'user_id': user_id,
            'after': after,
            'first': first
        }

        return self.request(Route(user_id, 'GET', 'moderation/channels', **params))

    def get_moderators(self,
                       broadcaster_id: str,
                       user_ids: Optional[List[str]] = None,
                       first: int = 20,
                       after: Optional[str] = None
                       ) -> Response[PData[List[users.SpecificUser]]]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'user_id': user_ids,
            'first': first,
            'after': after
        }
        return self.request(Route(broadcaster_id, 'GET', 'moderation/moderators', **params))

    def add_channel_moderator(self,
                              broadcaster_id: str,
                              user_id: str
                              ) -> Response[None]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'user_id': user_id
        }
        return self.request(Route(broadcaster_id, 'POST', 'moderation/moderators', **params))

    def remove_channel_moderator(self,
                                 broadcaster_id: str,
                                 user_id: str
                                 ) -> Response[None]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'user_id': user_id
        }
        return self.request(Route(broadcaster_id, 'DELETE', 'moderation/moderators', **params))

    def get_vips(self,
                 broadcaster_id: str,
                 user_ids: Optional[List[str]] = None,
                 first: int = 20,
                 after: Optional[str] = None
                 ) -> Response[PData[List[users.SpecificUser]]]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'user_id': user_ids,
            'first': first,
            'after': after
        }
        return self.request(Route(broadcaster_id, 'GET', 'channels/vips', **params))

    def add_channel_vip(self,
                        broadcaster_id: str,
                        user_id: str
                        ) -> Response[None]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'user_id': user_id
        }
        return self.request(Route(broadcaster_id, 'POST', 'channels/vips', **params))

    def remove_channel_vip(self,
                           broadcaster_id: str,
                           user_id: str
                           ) -> Response[None]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'user_id': user_id
        }
        return self.request(Route(broadcaster_id, 'DELETE', 'channels/vips', **params))

    def update_shield_mode_status(self,
                                  broadcaster_id: str,
                                  moderator_id: str,
                                  is_active: bool
                                  ) -> Response[Data[List[moderation.ShieldModeStatus]]]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'moderator_id': moderator_id
        }
        data: Dict[str, Any] = {
            'is_active': is_active
        }
        return self.request(Route(moderator_id, 'PUT', 'moderation/shield_mode', **params), data=data)

    def get_shield_mode_status(self,
                               broadcaster_id: str,
                               moderator_id: str
                               ) -> Response[Data[List[moderation.ShieldModeStatus]]]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'moderator_id': moderator_id
        }
        return self.request(Route(moderator_id, 'GET', 'moderation/shield_mode', **params))

    def warn_chat_user(self,
                       broadcaster_id: str,
                       moderator_id: str,
                       user_id: str,
                       reason: str
                       ) -> Response[Data[List[moderation.UserWarningResponse]]]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'moderator_id': moderator_id
        }
        body: Dict[str, Any] = {
            'user_id': user_id,
            'reason': reason
        }
        return self.request(Route(moderator_id, 'POST', 'moderation/warnings', **params), data=body)

    # Polls
    def get_polls(self,
                  broadcaster_id: str,
                  poll_ids: Optional[List[str]] = None,
                  first: int = 20,
                  after: Optional[str] = None
                  ) -> Response[PData[List[interaction.Poll]]]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'id': poll_ids,
            'first': first,
            'after': after
        }

        return self.request(Route(broadcaster_id, 'GET', 'polls', **params))

    def create_poll(self,
                    broadcaster_id: str,
                    title: str,
                    choices: List[str],
                    duration: int,
                    channel_points_voting_enabled: bool = False,
                    channel_points_per_vote: Optional[int] = None
                    ) -> Response[Data[List[interaction.Poll]]]:
        body: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'title': title,
            'choices': [{'title': choice} for choice in choices],
            'duration': duration,
            'channel_points_voting_enabled': channel_points_voting_enabled,
            'channel_points_per_vote': channel_points_per_vote
        }
        body = {key: value for key, value in body.items() if value is not None}
        return self.request(Route(broadcaster_id, 'POST', 'polls'), data=body)

    def end_poll(self,
                 broadcaster_id: str,
                 poll_id: str,
                 status: Literal['terminated', 'archived']
                 ) -> Response[Data[List[interaction.Poll]]]:
        body: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'id': poll_id,
            'status': status.upper()
        }
        return self.request(Route(broadcaster_id, 'PATCH', 'polls'), data=body)

    # Predictions
    def get_predictions(self,
                        broadcaster_id: str,
                        prediction_ids: Optional[List[str]] = None,
                        first: int = 20,
                        after: Optional[str] = None) -> Response[PData[List[interaction.Prediction]]]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'id': prediction_ids,
            'first': first,
            'after': after
        }
        return self.request(Route(broadcaster_id, 'GET', 'predictions', **params))

    def create_prediction(self,
                          broadcaster_id: str,
                          title: str,
                          outcomes: List[str],
                          prediction_window: int
                          ) -> Response[Data[List[interaction.Prediction]]]:
        body: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'title': title,
            'outcomes': [{'title': outcome} for outcome in outcomes],
            'prediction_window': prediction_window
        }
        return self.request(Route(broadcaster_id, 'POST', 'predictions'), data=body)

    def end_prediction(self,
                       broadcaster_id: str,
                       prediction_id: str,
                       status: Literal['resolved', 'canceled', 'locked'],
                       winning_outcome_id: Optional[str] = None
                       ) -> Response[Data[List[interaction.Prediction]]]:
        body: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'id': prediction_id,
            'status': status.upper(),
            'winning_outcome_id': winning_outcome_id
        }
        body = {key: value for key, value in body.items() if value is not None}
        return self.request(Route(broadcaster_id, 'PATCH', 'predictions'), data=body)

    # Raid
    def start_raid(self,
                   from_broadcaster_id: str,
                   to_broadcaster_id: str
                   ) -> Response[Data[List[streams.RaidInfo]]]:
        body: Dict[str, Any] = {
            'from_broadcaster_id': from_broadcaster_id,
            'to_broadcaster_id': to_broadcaster_id
        }
        return self.request(Route(from_broadcaster_id, 'POST', 'raids'), data=body)

    def cancel_raid(self,
                    broadcaster_id: str
                    ) -> Response[None]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id
        }
        return self.request(Route(broadcaster_id, 'DELETE', 'raids', **params))

    def get_channel_stream_schedule(self, __id: str,
                                    broadcaster_id: str,
                                    segment_ids: Optional[List[str]] = None,
                                    start_time: Optional[str] = None,
                                    first: int = 20,
                                    after: Optional[str] = None
                                    ) -> Response[PData[List[streams.Schedule]]]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'id': segment_ids,
            'start_time': start_time,
            'first': first,
            'after': after
        }

        return self.request(Route(__id, 'GET', 'schedule', **params))

    def get_channel_icalendar(self, __id: str, broadcaster_id: str) -> Response[str]:
        return self.request(Route(__id, 'GET', 'schedule/icalendar', broadcaster_id=broadcaster_id))

    def update_channel_stream_schedule(self,
                                       broadcaster_id: str,
                                       is_vacation_enabled: bool,
                                       vacation_start_time: Optional[str] = None,
                                       vacation_end_time: Optional[str] = None,
                                       timezone: Optional[str] = None
                                       ) -> Response[None]:
        params: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'is_vacation_enabled': is_vacation_enabled,
            'vacation_start_time': vacation_start_time,
            'vacation_end_time': vacation_end_time,
            'timezone': timezone
        }
        return self.request(Route(broadcaster_id, 'PATCH', 'schedule/settings', **params))

    def create_channel_schedule_segment(self,
                                        broadcaster_id: str,
                                        start_time: str,
                                        timezone: str,
                                        duration: int,
                                        is_recurring: Optional[bool] = None,
                                        category_id: Optional[str] = None,
                                        title: Optional[str] = None
                                        ) -> Response[Data[List[streams.Schedule]]]:
        body: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'start_time': start_time,
            'timezone': timezone,
            'duration': str(duration),
            'is_recurring': is_recurring,
            'category_id': category_id,
            'title': title
        }
        body = {key: value for key, value in body.items() if value is not None}
        return self.request(Route(broadcaster_id, 'POST', 'schedule/segment'), data=body)

    def update_channel_schedule_segment(self,
                                        broadcaster_id: str,
                                        segment_id: str,
                                        start_time: Optional[str] = None,
                                        duration: Optional[int] = None,
                                        category_id: Optional[str] = None,
                                        title: Optional[str] = None,
                                        is_canceled: Optional[bool] = None,
                                        timezone: Optional[str] = None
                                        ) -> Response[Data[List[streams.Schedule]]]:
        body: Dict[str, Any] = {
            'broadcaster_id': broadcaster_id,
            'id': segment_id,
            'start_time': start_time,
            'duration': str(duration),
            'category_id': category_id,
            'title': title,
            'is_canceled': is_canceled,
            'timezone': timezone
        }
        body = {key: value for key, value in body.items() if value is not None}
        return self.request(Route(broadcaster_id, 'PATCH', 'schedule/segment'), data=body)

    def delete_channel_schedule_segment(self,
                                        broadcaster_id: str,
                                        segment_id: str
                                        ) -> Response[None]:
        params = {
            'broadcaster_id': broadcaster_id,
            'id': segment_id
        }
        return self.request(Route(broadcaster_id, 'DELETE', 'schedule/segment', **params))

    # Search
    def search_categories(self, __id: str,
                          query: str,
                          first: int = 20,
                          after: Optional[str] = None
                          ) -> Response[PData[List[search.CategorySearch]]]:
        params = {
            'query': query,
            'first': first,
            'after': after
        }
        return self.request(Route(__id, 'GET', 'search/categories', **params))

    def search_channels(self, __id: str,
                        query: str,
                        live_only: bool = False,
                        first: int = 20,
                        after: Optional[str] = None
                        ) -> Response[PData[List[search.ChannelSearch]]]:
        params = {
            'query': query,
            'live_only': live_only,
            'first': first,
            'after': after
        }
        return self.request(Route(__id, 'GET', 'search/channels', **params))

    # Streams
    def get_stream_key(self, broadcaster_id: str) -> Response[Data[List[streams.StreamKey]]]:
        return self.request(Route(broadcaster_id, 'GET', 'streams/key', broadcaster_id=broadcaster_id))

    def get_streams(self, __id: str,
                    user_ids: Optional[List[str]] = None,
                    user_logins: Optional[List[str]] = None,
                    game_ids: Optional[List[str]] = None,
                    stream_type: Literal['all', 'live'] = 'all',
                    language: Optional[str] = None,
                    first: int = 20,
                    before: Optional[str] = None,
                    after: Optional[str] = None
                    ) -> Response[PData[List[streams.StreamInfo]]]:
        params = {
            'user_id': user_ids,
            'user_login': user_logins,
            'game_id': game_ids,
            'type': stream_type,
            'language': language,
            'first': first,
            'before': before,
            'after': after
        }
        return self.request(Route(__id, 'GET', 'streams', **params))

    def get_followed_streams(self,
                             user_id: str,
                             first: int = 100,
                             after: Optional[str] = None
                             ) -> Response[PData[List[streams.StreamInfo]]]:
        params = {
            'user_id': user_id,
            'first': first,
            'after': after
        }
        return self.request(Route(user_id, 'GET', 'streams/followed', **params))

    def create_stream_marker(self, __id: str,
                             user_id: str,
                             description: Optional[str] = None
                             ) -> Response[Data[List[streams.StreamMarkerInfo]]]:
        body: Dict[str, Any] = {
            'user_id': user_id,
            'description': description
        }
        return self.request(Route(__id, 'POST', 'streams/markers'), data=body)

    def get_stream_markers(self, __id: str,
                           user_id: [str] = None,
                           video_id: Optional[str] = None,
                           first: Optional[int] = 20,
                           before: Optional[str] = None,
                           after: Optional[str] = None
                           ) -> Response[PData[List[streams.StreamMarker]]]:
        params = {
            'user_id': user_id,
            'video_id': video_id,
            'first': first,
            'before': before,
            'after': after
        }
        return self.request(Route(__id, 'GET', 'streams/markers', **params))

    # Subscriptions
    def get_broadcaster_subscriptions(self,
                                      broadcaster_id: str,
                                      user_ids: Optional[List[str]] = None,
                                      first: Optional[int] = 20,
                                      before: Optional[str] = None,
                                      after: Optional[str] = None
                                      ) -> Response[TPData[List[channels.Subscription]]]:
        params = {
            'broadcaster_id': broadcaster_id,
            'user_id': user_ids,
            'first': first,
            'before': before,
            'after': after
        }
        return self.request(Route(broadcaster_id, 'GET', 'subscriptions', **params))

    def check_user_subscription(self,
                                user_id: str,
                                broadcaster_id: str
                                ) -> Response[Data[List[channels.SubscriptionCheck]]]:
        params = {
            'broadcaster_id': broadcaster_id,
            'user_id': user_id
        }
        return self.request(Route(user_id, 'GET', 'subscriptions/user', **params))

    def get_team_info(self, __id: str,
                      team_name: Optional[str] = None,
                      team_id: Optional[str] = None
                      ) -> Response[Data[List[channels.Team]]]:
        params = {
            'name': team_name.replace(' ', '').lower(),
            'id': team_id
        }
        return self.request(Route(__id, 'GET', 'teams', **params))

    def get_channel_teams(self, __id: str, broadcaster_id: str) -> Response[Data[List[channels.ChannelTeam]]]:
        return self.request(Route(__id, 'GET', 'teams/channel', broadcaster_id=broadcaster_id))

    # Users
    def get_users(self, __id: str,
                  user_ids: Optional[List[str]] = None,
                  user_logins: Optional[List[str]] = None) -> Response[Data[List[users.User]]]:
        params = {
            'login': user_logins,
            'id': user_ids
        }
        return self.request(Route(__id, 'GET', 'users', **params))

    def update_user(self, __id: str, description: str) -> Response[Data[List[users.User]]]:
        return self.request(Route(__id, 'PUT', 'users', description=description))

    def get_user_block_list(self,
                            broadcaster_id: str,
                            first: int = 20,
                            after: str = None
                            ) -> Response[PData[List[users.SpecificUser]]]:
        params = {
            'broadcaster_id': broadcaster_id,
            'first': first,
            'after': after
        }
        return self.request(Route(broadcaster_id, 'GET', 'users/blocks', **params))

    def block_user(self, __id: str, user_id: str,
                   source_context: Literal['chat', 'whisper'] = None,
                   reason: Literal['harassment', 'spam', 'other'] = None) -> Response[None]:
        params = {
            'target_user_id': user_id,
            'source_context': source_context,
            'reason': reason
        }
        return self.request(Route(__id, 'PUT', 'users/blocks', **params))

    def unblock_user(self, __id: str, user_id: str) -> Response[None]:
        return self.request(Route(__id, 'DELETE', 'users/blocks', user_id=user_id))

    def get_user_extensions(self, __id: str) -> Response[Data[List[channels.Extension]]]:
        return self.request(Route(__id, 'GET', 'users/extensions/list'))

    def get_user_active_extensions(self, __id: str,
                                   user_id: Optional[str] = None
                                   ) -> Response[Data[channels.ActiveExtensions]]:
        return self.request(Route(__id, 'GET', 'users/extensions', user_id=user_id))

    def update_user_extensions(self, __id: str,
                               key: Literal['panel', 'overlay', 'component'],
                               number: Literal['1', '2', '3'],
                               extension_id: str,
                               extension_version: str,
                               activate: bool,
                               x: Optional[int] = None,
                               y: Optional[int] = None) -> Response[Data[channels.ActiveExtensions]]:
        body = {
            'data': {str(key): {number: {'id': extension_id,
                                         'version': extension_version,
                                         'active': activate}}}
        }
        if x and y:
            body[key][number].update({'x': x, 'y': y})
        return self.request(Route(__id, 'PUT', 'users/extensions'), data=body)

    # Videos
    def get_videos(self, __id: str,
                   video_ids: Optional[List[str]] = None,
                   user_id: Optional[str] = None,
                   game_id: Optional[str] = None,
                   language: Optional[str] = None,
                   period: Optional[Literal['all', 'day', 'month', 'week']] = None,
                   sort: Optional[Literal['time', 'trending', 'views']] = None,
                   video_type: Optional[Literal['all', 'archive', 'highlight', 'upload']] = None,
                   first: Optional[int] = 20,
                   after: Optional[str] = None,
                   before: Optional[str] = None
                   ) -> Response[PData[List[channels.Video]]]:
        params: Dict[str, Any] = {
            'id': video_ids,
            'user_id': user_id,
            'game_id': game_id,
            'language': language,
            'period': period,
            'sort': sort,
            'type': video_type,
            'first': first,
            'after': after,
            'before': before
        }
        return self.request(Route(__id, 'GET', 'videos', **params))

    def delete_videos(self, __id: str, video_ids: List[str]) -> Response[Data[List[str]]]:
        return self.request(Route(__id, 'DELETE', 'videos', id=video_ids))

    # Whispers
    def send_whisper(self,
                     from_user_id: str,
                     to_user_id: str,
                     message: str
                     ) -> Response[None]:
        params: Dict[str, Any] = {
            'from_user_id': from_user_id,
            'to_user_id': to_user_id,
        }
        body: Dict[str, Any] = {
            'message': message
        }
        return self.request(Route(from_user_id, 'POST', 'whispers', **params), data=body)
